# import unittest
# import datetime
#
# from bson import ObjectId
#
# from mokito.fields import (
#     IntField,
#     FloatField,
#     StringField,
#     BooleanField,
#     ObjectIdField,
#     DateTimeField
# )
# from mokito.tools import make_field
#
#
# class TestFieldsDefault(unittest.TestCase):
#     def test_int_field_default(self):
#         f = make_field(123)
#         self.assertFalse(f.dirty)
#         self.assertIsInstance(f, IntField)
#         self.assertEqual(123, f.value)
#         self.assertIsNone(f.self_value)
#         f.set_value('456')
#         self.assertEqual(456, f.value)
#         self.assertEqual(456, f.self_value)
#         f.value = 789
#         self.assertEqual(789, f.value)
#         self.assertEqual(789, f.self_value)
#
#         with self.assertRaises(ValueError):
#             f.set_value('aaa')
#         self.assertEqual(789, f.value)
#
#     def test_float_field_const(self):
#         f = make_field(123.4)
#         self.assertIsInstance(f, FloatField)
#         self.assertEqual(f.value, 123.4)
#         self.assertIsNone(f.self_value)
#         f.value = 789
#         self.assertEqual(789.0, f.value)
#         self.assertEqual(789.0, f.self_value)
#         f.value = '456.7'
#         self.assertEqual(456.7, f.value)
#         self.assertEqual(456.7, f.self_value)
#
#     def test_str_field_const(self):
#         f = make_field('foo')
#         self.assertIsInstance(f, StringField)
#         self.assertEqual('foo', f.value)
#         self.assertIsNone(f.self_value)
#         f.value = 456.7
#         self.assertEqual('456.7', f.value)
#         self.assertEqual('456.7', f.self_value)
#
#     def test_bool_field_const(self):
#         f = make_field(True)
#         self.assertIsInstance(f, BooleanField)
#         self.assertTrue(f.value)
#         self.assertIsNone(f.self_value)
#         f.value = 123
#         self.assertTrue(f.value)
#         self.assertTrue(f.self_value)
#         f.value = None
#         self.assertTrue(f.value)
#         self.assertFalse(f.self_value)
#         f.value = True
#         self.assertTrue(f.value)
#         self.assertTrue(f.self_value)
#         f.value = False
#         self.assertFalse(f.value)
#         self.assertFalse(f.self_value)
#
#     def test_ObjectId_field_const(self):
#         id1 = ObjectId()
#         id2 = ObjectId()
#         f = make_field(id1)
#         self.assertIsInstance(f, ObjectIdField)
#         self.assertEqual(f.value, id1)
#         self.assertIsNone(f.self_value)
#         self.assertEqual(str(id1), f.get_value(_format='json'))
#         f.value = str(id2)
#         self.assertEqual(id2, f.value)
#         self.assertEqual(id2, f.self_value)
#         self.assertEqual(str(id2), f.get_value(_format='json'))
#         f.value = str(id1)
#         self.assertEqual(id1, f.value)
#         self.assertEqual(id1, f.self_value)
#         self.assertEqual(str(id1), f.get_value(_format='json'))
#
#     def test_datetime_field_const(self):
#         dt1 = datetime.datetime.now()
#         dt2 = datetime.datetime.utcnow()
#         f = make_field(dt1)
#         self.assertIsInstance(f, DateTimeField)
#         self.assertEqual(f.value, dt1)
#         self.assertIsNone(f.self_value)
#         self.assertEqual(dt1.replace(microsecond=0).isoformat(), f.get_value(_date_format='iso'))
#         self.assertEqual(dt1.strftime('%d/%m/%y'), f.get_value(_date_format='%d/%m/%y'))
#         f.value = dt2
#         self.assertEqual(dt2, f.value)
#         self.assertEqual(dt2, f.self_value)
#         self.assertEqual(dt2.replace(microsecond=0).isoformat(), f.get_value(_date_format='iso'))
#         self.assertEqual(dt2.strftime('%d/%m/%y'), f.get_value(_date_format='%d/%m/%y'))



# import unittest
# import datetime
#
# from bson import ObjectId
#
# from mokito.fields import (
#     IntField,
#     FloatField,
#     StringField,
#     BooleanField,
#     ObjectIdField,
#     DateTimeField,
#     ChoiceField
# )
# from mokito.tools import make_field
# from mokito.errors import MokitoChoiceError
#
#
# class TestFields(unittest.TestCase):
#
#     def test_int_field(self):
#         f = make_field(int)
#         self.assertIsInstance(f, IntField)
#         self.assertFalse(f.dirty)
#         self.assertIsNone(f.value)
#         self.assertIsNone(f.self_value)
#         f.set_value(123)
#         self.assertEqual(123, f.value)
#         self.assertEqual(123, f.self_value)
#         f.value = 789
#         self.assertEqual(789, f.value)
#         self.assertEqual(789, f.self_value)
#
#         with self.assertRaises(ValueError):
#             f.set_value('aaa')
#         self.assertEqual(789, f.value)
#
#     def test_float_field(self):
#         f = make_field(float)
#         self.assertIsInstance(f, FloatField)
#         self.assertIsNone(f.value)
#         self.assertIsNone(f.self_value)
#         f.set_value(123)
#         self.assertEqual(123.0, f.value)
#         self.assertEqual(123.0, f.self_value)
#         f.value = '456.7'
#         self.assertEqual(456.7, f.value)
#         self.assertEqual(456.7, f.self_value)
#
#     def test_str_field(self):
#         f = make_field(str)
#         self.assertIsInstance(f, StringField)
#         self.assertIsNone(f.value)
#         self.assertIsNone(f.self_value)
#         f.set_value(123)
#         self.assertEqual('123', f.value)
#         self.assertEqual('123', f.self_value)
#         f.value = b'456.7'
#         self.assertEqual('456.7', f.value)
#         self.assertEqual('456.7', f.self_value)
#
#     def test_bool_field(self):
#         f = make_field(bool)
#         self.assertIsInstance(f, BooleanField)
#         self.assertIsNone(f.value)
#         self.assertIsNone(f.self_value)
#         f.set_value(123)
#         self.assertTrue(f.value)
#         self.assertTrue(f.self_value)
#         f.value = None
#         self.assertFalse(f.value)
#         self.assertFalse(f.self_value)
#
#     def test_ObjectId_field(self):
#         id1 = ObjectId()
#         id2 = ObjectId()
#         f = make_field(ObjectId)
#         self.assertIsInstance(f, ObjectIdField)
#         self.assertIsNone(f.value)
#         self.assertIsNone(f.self_value)
#         self.assertIsNone(f.get_value(_format='json'))
#         f.set_value(id1)
#         self.assertEqual(id1, f.value)
#         self.assertEqual(id1, f.self_value)
#         self.assertEqual(str(id1), f.get_value(_format='json'))
#         f.value = str(id2)
#         self.assertEqual(id2, f.value)
#         self.assertEqual(id2, f.self_value)
#         self.assertEqual(str(id2), f.get_value(_format='json'))
#
#     def test_datetime_field(self):
#         dt1 = datetime.datetime.now()
#         dt2 = datetime.datetime.utcnow()
#         f = make_field(datetime.datetime)
#         self.assertIsInstance(f, DateTimeField)
#         self.assertIsNone(f.value)
#         self.assertIsNone(f.self_value)
#         self.assertIsNone(f.get_value(as_json=True))
#         f.set_value(dt1)
#         self.assertEqual(dt1, f.value)
#         self.assertEqual(dt1, f.self_value)
#         self.assertEqual(dt1.replace(microsecond=0).isoformat(), f.get_value(_date_format='iso'))
#         self.assertEqual(dt1.strftime('%d/%m/%y'), f.get_value(_date_format='%d/%m/%y'))
#         f.value = dt2
#         self.assertEqual(dt2, f.value)
#         self.assertEqual(dt2, f.self_value)
#         self.assertEqual(dt2.replace(microsecond=0).isoformat(), f.get_value(_date_format='iso'))
#         self.assertEqual(dt2.strftime('%d/%m/%y'), f.get_value(_date_format='%d/%m/%y'))
#         f.set_value('14/12/16', _date_format='%d/%m/%y')
#         self.assertEqual(datetime.datetime(2016, 12, 14), f.value)
#         self.assertEqual(datetime.datetime(2016, 12, 14), f.self_value)
#
#     def test_choice_field(self):
#         f = make_field(ChoiceField({1: 'a', 2: 'b'}))
#
#         with self.assertRaises(MokitoChoiceError):
#             f.value = 1
#
#         f.set_value(1, inner=True)
#         self.assertEqual(f.value, 'a')
#         self.assertEqual(f.get_value(inner=True), 1)
#
#         f.value = 'b'
#         self.assertEqual(f.value, 'b')
#         self.assertEqual(f.get_value(inner=True), 2)
#
#         f._val = 111
#         self.assertIsNone(f.value)



import datetime
try:
    import ujson as json
except ImportError:
    import json

import pytz
from bson import ObjectId
from dateutil.parser import parse

from .errors import MokitoChoiceError
from .tools import SEPARATOR


class Field(object):
    def __init__(self, _default=None, _parent=None, **kwargs):
        self._val = None
        self._parent = _parent
        self._dirty = False
        self._default = _default

    def __str__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.value)

    @property
    def parent(self):
        return self._parent

    def get_dirty(self):
        return self._dirty

    def set_dirty(self, value):
        self._dirty = value

    dirty = property(get_dirty, set_dirty)

    def dirty_clear(self):
        self.dirty = False

    def clear(self):
        self._val = self._default
        self.dirty = True

    def validate(self, value, **kwargs):
        return value

    def get_value(self, **kwargs):
        return self._default if self._val is None else self._val

    def set_value(self, value, **kwargs):
        value = self.validate(value, **kwargs)
        res = self._val != value
        if res:
            self._val = value
            self.dirty = True

    value = property(get_value, set_value)

    @property
    def self_value(self):
        return self._val


class AnyField(Field):
    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            if isinstance(self._val, (list, tuple)):
                k1 = int(k1)
            item = self._val[k1]
        except KeyError:
            item = AnyField(_parent=self)
        if k2:
            item = item.__getitem__(k2)

        return item if isinstance(item, Field) else AnyField(item)

    def validate(self, value, **kwargs):
        if isinstance(value, tuple):
            value = list(value)
        return value


class NumberField(Field):
    def __iadd__(self, other):
        if isinstance(other, Field):
            other = other.get_value()
        return (self._val or 0) + other


class IntField(NumberField):
    def validate(self, value, **kwargs):
        if value is not None:
            value = int(value)
        return value


class FloatField(NumberField):
    def validate(self, value, **kwargs):
        if value is not None:
            value = float(value)
        return value


class StringField(Field):
    def validate(self, value, **kwargs):
        if value is not None:
            if isinstance(value, (bytes, bytearray)):
                value = str(value, 'utf-8')
            elif not isinstance(value, str):
                value = str(value)
        return value


class BooleanField(Field):
    def validate(self, value, **kwargs):
        if value is not None:
            value = bool(value)
        return value


class ObjectIdField(Field):
    def validate(self, value, **kwargs):
        if value is not None:
            value = ObjectId(value)
        return value

    def get_value(self, _format=None, **kwargs):
        value = self._default if self._val is None else self._val
        if value is not None:
            return str(value) if _format == 'json' else value

    value = property(get_value, Field.set_value)


class DateTimeField(Field):
    def validate(self, value, _date_format=None, **kwargs):
        if not (value is None or isinstance(value, datetime.datetime)):
            if _date_format and _date_format != 'iso':
                value = datetime.datetime.strptime(value, _date_format)
            else:
                try:
                    value = parse(value).replace(tzinfo=None)
                except TypeError:
                    value = None
        return value

    def get_value(self, _date_format=None, tz_name=None, without_microsecond=True, tz=None, **kwargs):
        value = self._default if self._val is None else self._val
        if _date_format is None and (tz_name or tz):
            _date_format = 'iso'
        if value is None or _date_format is None:
            return value

        _tz = pytz.timezone(tz_name) if tz_name else tz
        val = _tz.fromutc(value) if _tz else value
        if without_microsecond:
            val = val.replace(microsecond=0)

        return val.isoformat() if _date_format.lower() == 'iso' else val.strftime(_date_format)

    value = property(get_value, Field.set_value)


class ChoiceField(Field):
    def __init__(self, choices, **kwargs):
        """
        :param choices: {mongo_value: orm_value} or [mongo_value] or (mongo_value,)
        :param kwargs:
        """
        super().__init__(**kwargs)
        if isinstance(choices, dict):
            self._choices = choices
        elif isinstance(choices, (list, tuple)):
            self._choices = {i: i for i in choices}
        else:
            raise TypeError()

    def _py_2_mongo(self, value):
        for k, v in self._choices.items():
            if value == v:
                return k

        if value is not None:
            raise MokitoChoiceError(value)

    def validate(self, value, inner=False, **kwargs):
        if not inner:
            value = self._py_2_mongo(value)
        return value

    def get_value(self, inner=False, **kwargs):
        value = self._default if self._val is None else self._val
        return value if inner else self._choices.get(value, None)

    value = property(get_value, Field.set_value)



# import unittest
#
# from mokito.tools import make_field
# from mokito.geo_fileds import (
#     GEOPointField,
#     GEOLineField,
#     GEOPolygonField,
#     GEOMultiPointField,
#     GEOMultiLineField
# )
# from mokito.models import Model
# from mokito.documents import Document
#
#
# class TestGeoFields(unittest.TestCase):
#     def test_point_field_1(self):
#         f = make_field(GEOPointField)
#         self.assertIsInstance(f, GEOPointField)
#         self.assertDictEqual(f.value, {
#             'type': 'Point',
#             'coordinates': [None, None]
#         })
#         self.assertDictEqual(f.query, {})
#         f['coordinates'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'Point',
#             'coordinates': [-1.0, -2.0]
#         })
#         self.assertEqual(f['coordinates.1'].value, -2.0)
#         f['coordinates.1'].value = 3.0
#         self.assertDictEqual(f.value, {
#             'type': 'Point',
#             'coordinates': [-1.0, 3.0]
#         })
#         with self.assertRaises(IndexError):
#             f['coordinates.3'].value = 4.0
#
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Point',
#             'coordinates': [-1.0, 3.0]}
#         })
#
#     def test_point_field_2(self):
#         f = make_field(GEOPointField([123, 45]))
#         self.assertIsInstance(f, GEOPointField)
#         self.assertDictEqual(f.value, {
#             'type': 'Point',
#             'coordinates': [123.0, 45.0]
#         })
#         self.assertDictEqual(f.query, {})
#         f['coordinates'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'Point',
#             'coordinates': [-1.0, -2.0]
#         })
#         self.assertEqual(f['coordinates.1'].value, -2.0)
#         f['coordinates.1'].value = 3.0
#         self.assertDictEqual(f.value, {
#             'type': 'Point',
#             'coordinates': [-1.0, 3.0]
#         })
#         with self.assertRaises(IndexError):
#             f['coordinates.3'].value = 4.0
#
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Point',
#             'coordinates': [-1.0, 3.0]}
#         })
#
#     def test_line_field_1(self):
#         f = make_field(GEOLineField)
#         self.assertIsInstance(f, GEOLineField)
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': []
#         })
#         self.assertDictEqual(f.query, {})
#
#         f['coordinates.0'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0]]
#         }})
#
#         f['coordinates'].append_value([3, 4])
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0], [3.0, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0], [3.0, 4.0]]
#         }})
#         self.assertEqual(f['coordinates.0.1'].value, -2.0)
#
#         f['coordinates.0.1'].value = 3.0
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0]]
#         }})
#
#         f['coordinates.3.1'].value = 4.0
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0]]
#         }})
#
#         with self.assertRaises(IndexError):
#             f['coordinates.3.4'].value = 5.0
#
#     def test_line_field_2(self):
#         f = make_field(GEOLineField([[1.23, 4.5]]))
#         self.assertIsInstance(f, GEOLineField)
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[1.23, 4.5]]
#         })
#         self.assertDictEqual(f.query, {})
#
#         f['coordinates.0'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0]]
#         }})
#
#         f['coordinates'].append_value([3, 4])
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0], [3.0, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0], [3.0, 4.0]]
#         }})
#         self.assertEqual(f['coordinates.0.1'].value, -2.0)
#
#         f['coordinates.0.1'].value = 3.0
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0]]
#         }})
#
#         f['coordinates.3.1'].value = 4.0
#         self.assertDictEqual(f.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0], [1.23, 4.5], [1.23, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0], [1.23, 4.5], [1.23, 4.0]]
#         }})
#
#         with self.assertRaises(IndexError):
#             f['coordinates.3.4'].value = 5.0
#
#     def test_polygon_field_1(self):
#         f = make_field(GEOPolygonField)
#         self.assertIsInstance(f, GEOPolygonField)
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': []
#         })
#         self.assertDictEqual(f.query, {})
#
#         f['coordinates.0.0'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, -2.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, -2.0]]
#             ]
#         }})
#
#         f['coordinates.0'].append_value([3, 4])
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, -2.0], [3.0, 4.0], [-1.0, -2.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, -2.0], [3.0, 4.0], [-1.0, -2.0]]
#             ]
#         }})
#         self.assertEqual(f['coordinates.0.0.1'].value, -2.0)
#
#         f['coordinates.0.0.1'].value = 3.0
#
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [-1.0, 3.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [-1.0, 3.0]]
#             ]
#         }})
#
#         f['coordinates.0.3.1'].value = 4.0
#
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [-1.0, 3.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [-1.0, 3.0]]
#             ]
#         }})
#
#         with self.assertRaises(IndexError):
#             f['coordinates.0.3.4'].value = 5.0
#
#         f['coordinates.0'].append_value([5, 6.0])
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0], [-1.0, 3.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0], [-1.0, 3.0]]
#             ]
#         }})
#
#         f['coordinates.1'].append_value([7, 8.0])
#         f['coordinates.1.1'].value = [9, 10.0]
#         self.assertDictEqual(f.value, {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0], [-1.0, 3.0]],
#                 [[7.0, 8.0], [9.0, 10.0], [7.0, 8.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0], [-1.0, 3.0]],
#                 [[7.0, 8.0], [9.0, 10.0], [7.0, 8.0]]
#             ]
#         }})
#
#     def test_multi_point_field_1(self):
#         f = make_field(GEOMultiPointField)
#         self.assertIsInstance(f, GEOMultiPointField)
#         self.assertDictEqual(f.value, {
#             'type': 'MultiPoint',
#             'coordinates': []
#         })
#         self.assertDictEqual(f.query, {})
#
#         f['coordinates.0'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, -2.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, -2.0]]
#         }})
#
#         f['coordinates'].append_value([3, 4])
#         self.assertDictEqual(f.value, {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, -2.0], [3.0, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, -2.0], [3.0, 4.0]]
#         }})
#         self.assertEqual(f['coordinates.0.1'].value, -2.0)
#
#         f['coordinates.0.1'].value = 3.0
#         self.assertDictEqual(f.value, {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0]]
#         }})
#
#         f['coordinates.3.1'].value = 4.0
#         self.assertDictEqual(f.value, {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0]]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiPoint',
#             'coordinates': [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0]]
#         }})
#
#         with self.assertRaises(IndexError):
#             f['coordinates.3.4'].value = 5.0
#
#     def test_multi_line_field_1(self):
#         f = make_field(GEOMultiLineField)
#         self.assertIsInstance(f, GEOMultiLineField)
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': []
#         })
#         self.assertDictEqual(f.query, {})
#
#         f['coordinates.0.0'].value = [-1, -2]
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, -2.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, -2.0]]
#             ]
#         }})
#
#         f['coordinates.0'].append_value([3, 4])
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, -2.0], [3.0, 4.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, -2.0], [3.0, 4.0]]
#             ]
#         }})
#         self.assertEqual(f['coordinates.0.0.1'].value, -2.0)
#
#         f['coordinates.0.0.1'].value = 3.0
#
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0]]
#             ]
#         }})
#
#         f['coordinates.0.3.1'].value = 4.0
#
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0]]
#             ]
#         }})
#
#         with self.assertRaises(IndexError):
#             f['coordinates.0.3.4'].value = 5.0
#
#         f['coordinates.0'].append_value([5, 6.0])
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0]]
#             ]
#         }})
#
#         f['coordinates.1'].append_value([7, 8.0])
#         f['coordinates.1.1'].value = [9, 10.0]
#         self.assertDictEqual(f.value, {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0]],
#                 [[7.0, 8.0], [9.0, 10.0]]
#             ]
#         })
#         self.assertDictEqual(f.query, {'$set': {
#             'type': 'MultiLineString',
#             'coordinates': [
#                 [[-1.0, 3.0], [3.0, 4.0], [None, None], [None, 4.0], [5.0, 6.0]],
#                 [[7.0, 8.0], [9.0, 10.0]]
#             ]
#         }})
#
#     def test_type_1(self):
#         f = make_field(GEOPointField)
#         self.assertEqual(f['type'].value, 'Point')
#         f['type'].value = 'foo'
#         self.assertEqual(f['type'].value, 'Point')
#
#     def test_set_line_field_1(self):
#         f1 = make_field(GEOLineField)
#         f2 = make_field(GEOLineField)
#         f1['coordinates.0'].value = [-1, -2]
#         f2['coordinates.0'] = f1['coordinates.0']
#         self.assertDictEqual(f2.value, {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0]]
#         })
#         self.assertDictEqual(f2.query, {'$set': {
#             'type': 'LineString',
#             'coordinates': [[-1.0, -2.0]]
#         }})
#
#     def test_set_polygon_field_1(self):
#         f1 = make_field(GEOPolygonField)
#         f2 = make_field(GEOPolygonField)
#         f1['coordinates.0'].value = [[-1, -2], [3, 4]]
#         f2['coordinates.0'] = f1['coordinates.0']
#         self.assertDictEqual(f2.value, {
#             'type': 'Polygon',
#             'coordinates': [[[-1.0, -2.0], [3.0, 4.0], [-1.0, -2.0]]]
#         })
#         self.assertDictEqual(f2.query, {'$set': {
#             'type': 'Polygon',
#             'coordinates': [[[-1.0, -2.0], [3.0, 4.0], [-1.0, -2.0]]]
#         }})
#
#     def test_model(self):
#         class GeoModel(Model):
#             scheme = {
#                 'a': GEOPointField([1.2, -3.4])
#             }
#
#         class GeoDocument(Document):
#             __collection__ = 'geo'
#             scheme = {
#                 'foo': GeoModel,
#                 'bar': GEOPolygonField
#             }
#
#         f = GeoDocument()
#         self.assertDictEqual(f.value, {
#             '_id': None,
#             'foo': {
#                 'a': {
#                     'type': 'Point',
#                     'coordinates': [1.2, -3.4]
#                 }
#             },
#             'bar': {
#                 'type': 'Polygon',
#                 'coordinates': []
#             }
#         })
#
#         f['foo.a.coordinates.1'].value = 5
#         f['bar.coordinates.0'].value = [[5, 6], [7, 8]]
#         self.assertDictEqual(f.value, {
#             '_id': None,
#             'foo': {
#                 'a': {
#                     'type': 'Point',
#                     'coordinates': [1.2, 5.0]
#                 }
#             },
#             'bar': {
#                 'type': 'Polygon',
#                 'coordinates': [
#                     [[5.0, 6.0], [7.0, 8.0], [5.0, 6.0]]
#                 ]
#             }
#         })



# import copy
#
# from bson import ObjectId
# from bson.errors import InvalidId
# from tornado.ioloop import IOLoop
# from tornado.testing import AsyncTestCase, gen_test
# from tornado.gen import coroutine
# import motor.motor_tornado
#
# import mokito
#
# import tests.util as u
#
#
# class ORMTestCase(AsyncTestCase):
#     def get_new_ioloop(self):
#         return IOLoop.instance()
#
#     @coroutine
#     def _init_collection(self):
#         yield self.db[u.TEST_COLLECTION1].insert_many([u.col1_data1, u.col1_data2])
#         yield self.db[u.TEST_COLLECTION2].insert_many([u.col2_data1, u.col2_data2])
#         yield self.db[u.TEST_COLLECTION3].insert_many([u.col3_data1, u.col3_data2])
#
#     @coroutine
#     def _drop_collection(self):
#         yield self.db.drop_collection(u.TEST_COLLECTION3)
#         yield self.db.drop_collection(u.TEST_COLLECTION2)
#         yield self.db.drop_collection(u.TEST_COLLECTION1)
#
#     def setUp(self):
#         super(ORMTestCase, self).setUp()
#         self.setUpHelper()
#
#     @gen_test
#     def setUpHelper(self):
#         self.db = motor.motor_tornado.MotorClient(u.TEST_DB_URI)[u.TEST_DB_NAME]
#         yield self._drop_collection()
#         yield self._init_collection()
#
#     def tearDown(self):
#         yield self.tearDownHelper()
#         super(ORMTestCase, self).tearDown()
#
#     @gen_test
#     def tearDownHelper(self):
#         yield self._drop_collection()
#
#     @gen_test
#     def test_find_one_1(self):
#         x = yield u.Document1.find_one(u.col1_id1)
#         self.assertEqual(x.id, u.col1_id1)
#
#         data = copy.deepcopy(u.col1_data1)
#         data['x3'][1] = 1
#         self.assertDictEqual(data, x.value)
#
#         x = yield u.Document1.find_one(u.col1_id2)
#         self.assertEqual(x.id, u.col1_id2)
#         data = copy.deepcopy(u.col1_data2)
#         data.pop('foo')
#         data['x3'][1] = 2
#         self.assertDictEqual(data, x.value)
#
#     @gen_test
#     def test_find_one_2(self):
#         _id = ObjectId()
#         x = yield u.Document1.find_one(_id)
#         self.assertIsNone(x)
#
#         with self.assertRaises(InvalidId):
#             yield u.Document1.find_one('foo')
#
#     @gen_test
#     async def test_find_1(self):
#         x = await u.Document1.find().all()
#         self.assertEqual(len(x), 2)
#
#         data = copy.deepcopy(u.col1_data1)
#         data['x3'][1] = 1
#         self.assertDictEqual(data, x[0].value)
#
#         data = copy.deepcopy(u.col1_data2)
#         data.pop('foo')
#         data['x3'][1] = 2
#         self.assertDictEqual(data, x[1].value)
#
#     @gen_test
#     async def test_find_2(self):
#         x = []
#         async for i in u.Document1.find():
#             x.append(i)
#         self.assertEqual(len(x), 2)
#
#         data = copy.deepcopy(u.col1_data1)
#         data['x3'][1] = 1
#         self.assertDictEqual(data, x[0].value)
#
#         data = copy.deepcopy(u.col1_data2)
#         data.pop('foo')
#         data['x3'][1] = 2
#         self.assertDictEqual(data, x[1].value)
#
#     @gen_test
#     def test_count(self):
#         self.assertEqual((yield u.Document1.count()), 2)
#
#     @gen_test
#     def test_save_1(self):
#         x = yield u.Document1.find_one(u.col1_id2)
#         self.assertFalse(x.dirty)
#         x['x1'].value = 3
#         self.assertTrue(x.dirty)
#         self.assertDictEqual(x.query, {'$set': {'x1': 3.0}})
#         y = yield x.save()
#         self.assertTrue(y)
#         self.assertFalse(x.dirty)
#
#         data = yield self.db[u.TEST_COLLECTION1].find_one(u.col1_id2, ['x1'])
#         self.assertEqual(data['x1'], 3.0)
#         self.assertEqual(x['x1'].value, 3.0)
#
#         data = yield self.db[u.TEST_COLLECTION1].count()
#         self.assertEqual(data, 2)
#
#     @gen_test
#     def test_save_2(self):
#         data = copy.deepcopy(u.col1_data1)
#         del data['_id']
#
#         x = u.Document1(data, inner=True)
#         self.assertIsNone(x.id)
#         self.assertTrue(x.dirty)
#         data['x3'][1] = 1
#         data['_id'] = None
#         self.assertDictEqual(x.value, data)
#
#         y = yield x.save()
#         self.assertTrue(y)
#         self.assertIsNotNone(x.id)
#         self.assertFalse(x.dirty)
#
#         data = yield self.db[u.TEST_COLLECTION1].find_one(x.id)
#         data['x3'][1] = 1
#         self.assertDictEqual(x.value, data)
#         self.assertEqual((yield self.db[u.TEST_COLLECTION1].count()), 3)
#
#     @gen_test
#     def test_save_3(self):
#         x = u.Document2()
#
#         self.assertIsNone(x.id)
#         self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 2)
#
#         res = yield x.save()
#         self.assertTrue(res)
#         self.assertDictEqual(x.value, {
#             '_id': x.id,
#             'f1': [None, None],
#             'f2': {'a': None, 'b': None},
#             'm1': {'x1': None, 'x2': [], 'x3': [None, None], 'x4': {'a': None, 'b': None}},
#             'm2': [],
#             'd1': {'x1': None, 'x2': [], 'x3': [None, None], 'x4': {'a': None, 'b': None}, '_id': None},
#             'd2': []
#         })
#         self.assertIsNotNone(x.id)
#         self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 3)
#
#         res = yield self.db[u.TEST_COLLECTION2].find_one({'_id': {'$nin': [u.col2_id1, u.col2_id2]}})
#         self.assertDictEqual(res, {
#             '_id': x.id,
#             'f1': [None, None],
#             'f2': {'a': None, 'b': None},
#             'm1': {'x1': None, 'x2': [], 'x3': [None, None], 'x4': {'a': None, 'b': None}},
#             'm2': [],
#             'd1': None,
#             'd2': []
#         })
#
#     @gen_test
#     def test_save_4(self):
#         x = u.Document2()
#         x['f1.1'].value = 'foo'
#         x['f2.a'].value = 123
#
#         self.assertIsNone(x.id)
#         self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 2)
#
#         res = yield x.save()
#         self.assertTrue(res)
#         self.assertDictEqual(x.value, {
#             '_id': x.id,
#             'f1': [None, 'foo'],
#             'f2': {'a': 123, 'b': None},
#             'm1': {'x1': None, 'x2': [], 'x3': [None, None], 'x4': {'a': None, 'b': None}},
#             'm2': [], 'd1': {'x1': None, 'x2': [], 'x3': [None, None], 'x4': {'a': None, 'b': None}, '_id': None},
#             # 'd1': {'_id': None, 'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
#             'd2': []
#         })
#         self.assertIsNotNone(x.id)
#         self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 3)
#
#         res = yield self.db[u.TEST_COLLECTION2].find_one({'_id': {'$nin': [u.col2_id1, u.col2_id2]}})
#         self.assertDictEqual(res, {
#             '_id': x.id,
#             'f1': [None, 'foo'],
#             'f2': {'a': 123, 'b': None},
#             'm1': {'x1': None, 'x2': [], 'x3': [None, None], 'x4': {'a': None, 'b': None}},
#             'm2': [],
#             'd1': None,
#             'd2': []
#         })
#
#     @gen_test
#     def test_load_1(self):
#         d = yield u.Document2.find_one(u.col2_id1)
#         self.assertDictEqual(d['d1'].value, {
#             '_id': u.col1_id1,
#             'x2': [],
#             'x3': [None, None],
#             'x1': None,
#             'x4': {'a': None, 'b': None}
#         })
#         self.assertEqual(d['d1'].dbref, u.col1_dbref1)
#
#         yield d['d1'].reload()
#         self.assertDictEqual(d['d1'].value, {
#             '_id': u.col1_id1,
#             'x1': 0.5,
#             'x2': [u.dt1, u.dt2, u.dt3],
#             'x3': [123, 1],
#             'x4': {'a': 1, 'b': 2}
#         })
#
#     @gen_test
#     def test_load_2(self):
#         d = yield u.Document2.find().all()
#
#         self.assertDictEqual(d['0.d1'].value, {
#             '_id': u.col1_id1,
#             'x2': [],
#             'x3': [None, None],
#             'x1': None,
#             'x4': {'a': None, 'b': None}
#         })
#         self.assertEqual(d['0.d1'].dbref, u.col1_dbref1)
#         yield d['0.d1'].reload()
#         self.assertDictEqual(d['0.d1'].value, {
#             '_id': u.col1_id1,
#             'x1': 0.5,
#             'x2': [u.dt1, u.dt2, u.dt3],
#             'x3': [123, 1],
#             'x4': {'a': 1, 'b': 2}
#         })
#
#     @gen_test
#     def test_to_prop(self):
#         x = yield u.Document1.find_one(u.col1_id1)
#         self.assertEqual(x.prop1, x['x1'].value + 2)
#         self.assertEqual(x.prop2, x['x1'].value * 2)
#         x['x1'].value = 10
#         self.assertEqual(x.prop1, 12.0)
#         self.assertEqual(x.prop2, 20.0)
#
#     @gen_test
#     def test_to_json_1(self):
#         d = yield u.Document1.find_one(u.col1_id1)
#
#         data = d.get_value(_date_format='iso', _format='json')
#         self.assertDictEqual(data, {
#             '_id': u.col1_id1,
#             'x1': 0.5,
#             'x2': ["2016-01-02T03:04:05", "2016-02-03T04:05:06", "2016-03-04T05:06:07"],
#             'x3': [123, 1],
#             'x4': {"a": 1, "b": 2}
#         })
#
#         data = d.get_value(_date_format='iso', tz_name='Asia/Novosibirsk')
#         self.assertDictEqual(data, {
#             '_id': u.col1_id1,
#             'x1': 0.5,
#             'x2': ['2016-01-02T09:04:05+06:00', '2016-02-03T10:05:06+06:00', '2016-03-04T11:06:07+06:00'],
#             'x3': [123, 1],
#             'x4': {"a": 1, "b": 2}
#         })
#
#     @gen_test
#     def test_to_json_2(self):
#         d = yield u.Document1.find().all()
#
#         data = d.get_value(_date_format='iso', _format='json')
#         self.assertListEqual(data, [
#             {
#                 '_id': u.col1_id1,
#                 'x1': 0.5,
#                 'x2': ['2016-01-02T03:04:05', '2016-02-03T04:05:06', '2016-03-04T05:06:07'],
#                 'x3': [123, 1],
#                 'x4': {'a': 1, 'b': 2}
#             }, {
#                 '_id': u.col1_id2,
#                 'x1': 1.5,
#                 'x2': ['2016-04-05T06:07:08', '2016-05-06T07:08:09', '2016-06-07T08:09:10'],
#                 'x3': [45601, 2],
#                 'x4': {'a': 3, 'b': 4}
#             }
#         ])
#
#     @gen_test
#     def test_from_json_1(self):
#         d = u.Document2()
#         self.assertFalse(d.dirty)
#         d.set_value(u.col2_data2, inner=True)
#         self.assertTrue(d.dirty)
#         self.assertEqual(d['d1'].dbref, u.col1_dbref2)
#
#         yield d['d1'].reload()
#         yield [i.reload() for i in d['d2']]
#         self.assertDictEqual(d.value, {
#             '_id': u.col2_id2,
#             'f1': [45602, 'foo3'],
#             'f2': {'a': 11, 'b': 12},
#             'm1': {
#                 'x1': 0.2,
#                 'x2': [u.dt11, u.dt12],
#                 'x3': [7, 3],
#                 'x4': {'a': 8, 'b': 8}
#             },
#             'm2': [
#                 {'x1': 0.2, 'x2': [u.xt1], 'x3': [8, 1], 'x4': {'a': 9, 'b': 9}},
#                 {'x1': 0.3, 'x2': [u.xt2, u.xt3], 'x3': [9, 2], 'x4': {'a': 10, 'b': 10}}],
#             'd1': {'x1': 1.5, 'x2': [u.dt4, u.dt5, u.dt6], 'x3': [45601, 2], 'x4': {'a': 3, 'b': 4}, '_id': u.col1_id2},
#             'd2': [{'x1': 1.5, 'x2': [u.dt4, u.dt5, u.dt6], 'x3': [45601, 2], 'x4': {'a': 3, 'b': 4}, '_id': u.col1_id2}]
#         })
#
#     @gen_test
#     def test_distinct(self):
#         data = yield u.Document2.distinct('m1.x1')
#         self.assertListEqual(data, [0.1, 0.2])
#
#     @gen_test
#     def test_set_document(self):
#         x1 = yield u.Document1.find_one(u.col1_id2)
#         x2 = yield u.Document2.find_one(u.col2_id1)
#
#         yield x2.reload('d1')
#         self.assertDictEqual(x2['d1'].value, {
#             '_id': u.col1_id1,
#             'x1': 0.5,
#             'x2': [u.dt1, u.dt2, u.dt3],
#             'x3': [123, 1],
#             'x4': {'a': 1, 'b': 2}
#         })
#
#         x2['d1'] = x1
#         self.assertTrue(x2.dirty)
#         self.assertDictEqual(x2.query, {'$set': {'d1': u.col1_dbref2}})
#
#         yield x2.reload('d1')
#         self.assertDictEqual(x2['d1'].value, {
#             '_id': u.col1_id2,
#             'x1': 1.5,
#             'x2': [u.dt4, u.dt5, u.dt6],
#             'x3': [45601, 2],
#             'x4': {'a': 3, 'b': 4}
#         })
#
#     @gen_test
#     def test_documents_1(self):
#         x1 = yield u.Document1.find().all()
#         self.assertEqual(len(x1), 2)
#         for i in x1:
#             self.assertIn(i.id, [u.col1_id1, u.col1_id2])
#
#         yield x1.remove()
#
#         x2 = yield u.Document1.find().all()
#         self.assertEqual(len(x2), 0)
#
#         yield x1.save()
#
#         x2 = yield u.Document1.find().all()
#         self.assertEqual(len(x2), 2)
#
#     @gen_test
#     def test_sort_1(self):
#         x = [i['foo'].value for i in (yield u.Document3.find().all())]
#         self.assertListEqual(x, [1, 2])
#
#         x = [i['foo'].value for i in (yield u.Document4.find().all())]
#         self.assertListEqual(x, [2, 1])
#
#         x = [i['foo'].value for i in (yield u.Document4.find(sort='foo').all())]
#         self.assertListEqual(x, [1, 2])
#
#         x = [i['foo'].value for i in (yield u.Document4.find(sort=['-foo']).all())]
#         self.assertListEqual(x, [2, 1])



import unittest
import datetime

from bson import ObjectId

from mokito.fields import (
    IntField,
    FloatField,
    StringField,
    BooleanField,
    ObjectIdField,
    DateTimeField,
    ChoiceField
)
from mokito.tools import make_field
from mokito.collections import ListField, TupleField, DictField


class TestCollectionFields(unittest.TestCase):
    # def test_list_field_1(self):
    #     f = make_field([int])
    #     self.assertIsInstance(f, ListField)
    #     self.assertIsInstance(f._rules, IntField)
    #
    #     f = make_field([float])
    #     self.assertIsInstance(f, ListField)
    #     self.assertIsInstance(f._rules, FloatField)
    #
    #     f = make_field([str])
    #     self.assertIsInstance(f, ListField)
    #     self.assertIsInstance(f._rules, StringField)
    #
    #     f = make_field([bool])
    #     self.assertIsInstance(f, ListField)
    #     self.assertIsInstance(f._rules, BooleanField)
    #
    #     f = make_field([datetime.datetime])
    #     self.assertIsInstance(f, ListField)
    #     self.assertIsInstance(f._rules, DateTimeField)
    #
    #     f = make_field([ObjectId])
    #     self.assertIsInstance(f, ListField)
    #     self.assertIsInstance(f._rules, ObjectIdField)
    #
    # def test_list_field_2(self):
    #     f = make_field([str])
    #     for i in [0, '0']:
    #         x = f[i]
    #         self.assertIsInstance(x, StringField)
    #         self.assertIsNone(x.value)
    #
    #     with self.assertRaises(TypeError):
    #         _ = f['foo']
    #
    #     with self.assertRaises(IndexError):
    #         _ = f['0.0']
    #
    # def test_list_field_3(self):
    #     f = make_field([str])
    #     self.assertEqual(f.value, [])
    #     self.assertEqual(f.self_value, [])
    #
    #     f[1].value = 'foo'
    #     self.assertEqual(f.value, [None, 'foo'])
    #     self.assertEqual(f.self_value, [None, 'foo'])
    #     self.assertTrue(f == f[1]._parent)
    #
    #     f.append_value('bar')
    #     self.assertEqual(f.value, [None, 'foo', 'bar'])
    #     self.assertEqual(f.self_value, [None, 'foo', 'bar'])
    #     self.assertTrue(f == f[2]._parent)
    #
    #     f.append(f[2])
    #     self.assertEqual(f.value, [None, 'foo', 'bar', 'bar'])
    #     self.assertEqual(f.self_value, [None, 'foo', 'bar', 'bar'])
    #     self.assertTrue(f == f[3]._parent)
    #
    #     with self.assertRaises(TypeError):
    #         f.append('xxx')
    #
    #     f['5'] = f[1]
    #     self.assertEqual(f.value, [None, 'foo', 'bar', 'bar', None, 'foo'])
    #     self.assertEqual(f.self_value, [None, 'foo', 'bar', 'bar', None, 'foo'])
    #     self.assertTrue(f == f[5]._parent)
    #
    #     self.assertIsNone(f.pop(0).value)
    #     self.assertEqual(f.value, ['foo', 'bar', 'bar', None, 'foo'])
    #     self.assertEqual(f.self_value, ['foo', 'bar', 'bar', None, 'foo'])
    #
    #     self.assertEqual(f.pop().get_value(), 'foo')
    #     self.assertEqual(f.value, ['foo', 'bar', 'bar'])
    #     self.assertEqual(f.self_value, ['foo', 'bar', 'bar'])
    #
    #     self.assertEqual(f.pop(5).value, None)
    #     self.assertEqual(f.value, ['foo', 'bar', 'bar'])
    #     self.assertEqual(f.self_value, ['foo', 'bar', 'bar'])
    #
    #     del f[0]
    #     self.assertEqual(f.value, ['bar', 'bar'])
    #     self.assertEqual(f.self_value, ['bar', 'bar'])
    #
    #     f.set_value(['x', 'bar', 'y'])
    #     self.assertEqual(f.value, ['bar', 'bar', 'y'])
    #     f.clear()
    #     f.set_value(['x', 'bar', 'y'])
    #     self.assertEqual(f.value, ['x', 'bar', 'y'])
    #     self.assertEqual(f.self_value, ['x', 'bar', 'y'])
    #
    #     f[1].set_value(None)
    #     self.assertEqual(f.value, ['x', None, 'y'])
    #     self.assertEqual(f.self_value, ['x', None, 'y'])
    #
    #     f.value = None
    #     self.assertEqual(f.value, [])
    #     self.assertEqual(f.self_value, [])
    #
    # def test_list_field_4(self):
    #     dt1 = datetime.datetime.utcnow()
    #     dt2 = datetime.datetime(2016, 1, 2, 3, 4, 5)
    #
    #     f = make_field([datetime.datetime])
    #     self.assertEqual(f.value, [])
    #     self.assertEqual(f.self_value, [])
    #
    #     f.set_value([dt1, dt2])
    #     self.assertEqual(f.value, [dt1, dt2])
    #     self.assertEqual(f.self_value, [dt1, dt2])
    #     self.assertEqual(f.get_value(_date_format='iso'),
    #                      [dt1.replace(microsecond=0).isoformat(),
    #                       dt2.replace(microsecond=0).isoformat()])
    #     self.assertEqual(f.get_value(_date_format='%d/%m/%y'),
    #                      [dt1.strftime('%d/%m/%y'), dt2.strftime('%d/%m/%y')])
    #     self.assertTrue(f == f[0]._parent)
    #     self.assertTrue(f == f[1]._parent)
    #
    # def test_tuple_field_1(self):
    #     f = make_field((int, str))
    #     self.assertIsInstance(f, TupleField)
    #     self.assertIsInstance(f._val['0'], IntField)
    #     self.assertTrue(f == f._val['0']._parent)
    #     self.assertIsInstance(f._val['1'], StringField)
    #     self.assertTrue(f == f._val['1']._parent)
    #     self.assertListEqual(f.value, [None, None])
    #
    #     for i in [0, '0']:
    #         self.assertIsNone(f[i].value)
    #         self.assertIsNone(f[i].self_value)
    #
    #     with self.assertRaises(TypeError):
    #         _ = f['a']
    #
    #     for i in ['0.0', 2]:
    #         with self.assertRaises(IndexError):
    #             _ = f[i]
    #
    # def test_tuple_field_2(self):
    #     f = make_field((datetime.datetime, str))
    #
    #     self.assertEqual(f.value, [None, None])
    #     self.assertEqual(f.self_value, [None, None])
    #     self.assertEqual(f.get_value(_date_format='iso'), [None, None])
    #
    #     f[1].value = 123
    #     self.assertEqual(f.value, [None, '123'])
    #     self.assertEqual(f.self_value, [None, '123'])
    #     self.assertTrue(f == f[1]._parent)
    #
    #     dt1 = datetime.datetime(2016, 1, 2, 3, 4, 5)
    #     f[0].value = dt1
    #     self.assertEqual(f.value, [dt1, '123'])
    #     self.assertEqual(f.self_value, [dt1, '123'])
    #     self.assertEqual(f.get_value(_date_format='iso'), ['2016-01-02T03:04:05', '123'])
    #     self.assertTrue(f == f[0]._parent)
    #
    #     f[1].value = None
    #     self.assertEqual(f.value, [dt1, None])
    #     self.assertEqual(f.self_value, [dt1, None])
    #     self.assertTrue(f == f[1]._parent)
    #
    #     del f[0]
    #     self.assertEqual(f.value, [None, None])
    #     self.assertEqual(f.self_value, [None, None])
    #     self.assertTrue(f == f[0]._parent)
    #
    #     f.set_value([dt1, 456])
    #     self.assertEqual(f.value, [dt1, '456'])
    #     self.assertEqual(f.self_value, [dt1, '456'])
    #     self.assertEqual(f.get_value(_date_format='%d/%m/%y'), [dt1.strftime('%d/%m/%y'), '456'])
    #     self.assertTrue(f == f[0]._parent)
    #     self.assertTrue(f == f[1]._parent)
    #
    # def test_dict_field_1(self):
    #     dt1 = datetime.datetime(2016, 1, 2, 3, 4, 5)
    #     dt2 = datetime.datetime(2016, 2, 3, 4, 5, 6)
    #     id1 = ObjectId()
    #     data = {
    #         'd': {
    #             'd1': datetime.datetime,
    #             'd2': [str],
    #             'd3': (ObjectId, str)
    #         },
    #         'l': [{'l1': int, 'l2': str}],
    #         't': (int, {'t1': datetime.datetime, 't2': str}),
    #     }
    #     f = make_field(data)
    #
    #     self.assertIsInstance(f, DictField)
    #     self.assertDictEqual(f.value, {
    #         'd': {'d1': None, 'd2': [], 'd3': [None, None]},
    #         'l': [],
    #         't': [None, {'t1': None, 't2': None}],
    #     })
    #     self.assertDictEqual(f.self_value, {
    #         'd': {'d1': None, 'd2': [], 'd3': [None, None]},
    #         'l': [],
    #         't': [None, {'t1': None, 't2': None}],
    #     })
    #     self.assertTrue(f == f['d']._parent)
    #     self.assertTrue(f == f['d']['d1']._parent._parent)
    #     self.assertTrue(f == f['d']['d2']._parent._parent)
    #     self.assertTrue(f == f['d']['d3']._parent._parent)
    #     self.assertTrue(f == f['d']['d3'][0]._parent._parent._parent)
    #     self.assertTrue(f == f['d']['d3'][1]._parent._parent._parent)
    #     self.assertTrue(f == f['l']._parent)
    #     self.assertTrue(f == f['t']._parent)
    #     self.assertTrue(f == f['t'][0]._parent._parent)
    #     self.assertTrue(f == f['t'][1]._parent._parent)
    #
    #     self.assertIsInstance(f._val['d'], DictField)
    #     self.assertIsInstance(f['d'], DictField)
    #     self.assertDictEqual(f['d'].value, {'d1': None, 'd2': [], 'd3': [None, None]})
    #     self.assertDictEqual(f['d'].self_value, {'d1': None, 'd2': [], 'd3': [None, None]})
    #
    #     self.assertIsInstance(f._val['l'], ListField)
    #     self.assertIsInstance(f['l'], ListField)
    #     self.assertEqual(f['l'].value, [])
    #     self.assertEqual(f['l'].self_value, [])
    #
    #     self.assertIsInstance(f._val['t'], TupleField)
    #     self.assertIsInstance(f['t'], TupleField)
    #     self.assertEqual(f['t'].value, [None, {'t1': None, 't2': None}])
    #     self.assertEqual(f['t'].self_value, [None, {'t1': None, 't2': None}])
    #
    #     f.set_value({
    #         'd': {
    #             'd1': dt1,
    #             'd2': ['foo'],
    #             'd3': (id1, 'bar')
    #         },
    #         'l': [{'l1': 456, 'l2': 'baz'}],
    #         't': (789, {'t1': dt2, 't2': 'x'})
    #     })
    #     self.assertTrue(f == f['d']._parent)
    #     self.assertTrue(f == f['d']['d1']._parent._parent)
    #     self.assertTrue(f == f['d']['d2']._parent._parent)
    #     self.assertTrue(f == f['d']['d3']._parent._parent)
    #     self.assertTrue(f == f['d']['d3'][0]._parent._parent._parent)
    #     self.assertTrue(f == f['d']['d3'][1]._parent._parent._parent)
    #     self.assertTrue(f == f['l']._parent)
    #     self.assertTrue(f == f['l'][0]._parent._parent)
    #     self.assertTrue(f == f['l'][0]['l1']._parent._parent._parent)
    #     self.assertTrue(f == f['l'][0]['l2']._parent._parent._parent)
    #     self.assertTrue(f == f['t']._parent)
    #     self.assertTrue(f == f['t'][0]._parent._parent)
    #     self.assertTrue(f == f['t'][1]._parent._parent)
    #     self.assertTrue(f == f['t'][1]['t1']._parent._parent._parent)
    #     self.assertTrue(f == f['t'][1]['t2']._parent._parent._parent)
    #
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': dt1,
    #             'd2': ['foo'],
    #             'd3': [id1, 'bar']
    #         },
    #         'l': [{'l1': 456, 'l2': 'baz'}],
    #         't': [789, {'t1': dt2, 't2': 'x'}]
    #     })
    #     self.assertDictEqual(f.self_value, {
    #         'd': {
    #             'd1': dt1,
    #             'd2': ['foo'],
    #             'd3': [id1, 'bar']
    #         },
    #         'l': [{'l1': 456, 'l2': 'baz'}],
    #         't': [789, {'t1': dt2, 't2': 'x'}]
    #     })
    #     self.assertDictEqual(f.get_value(_date_format='iso', _format='json'), {
    #         'd': {
    #             'd1': dt1.replace(microsecond=0).isoformat(),
    #             'd2': ['foo'],
    #             'd3': [str(id1), 'bar']
    #         },
    #         'l': [{'l1': 456, 'l2': 'baz'}],
    #         't': [789, {'t1': dt2.replace(microsecond=0).isoformat(), 't2': 'x'}],
    #     })
    #
    #     self.assertDictEqual(f['d'].value, {'d1': dt1, 'd2': ['foo'], 'd3': [id1, 'bar']})
    #     self.assertDictEqual(f['d'].self_value, {'d1': dt1, 'd2': ['foo'], 'd3': [id1, 'bar']})
    #
    #     self.assertEqual(f['l'].value, [{'l2': 'baz', 'l1': 456}])
    #     self.assertEqual(f['l'].self_value, [{'l2': 'baz', 'l1': 456}])
    #
    #     self.assertEqual(f['t'].value, [789, {'t1': dt2, 't2': 'x'}])
    #     self.assertEqual(f['t'].self_value, [789, {'t1': dt2, 't2': 'x'}])
    #
    #     f['d.d1'].value = dt2
    #     f['d.d2'].append_value('abc')
    #     f['d.d3.1'].value = 'xyz'
    #     f['l'].append_value({'l1': 100, 'l2': 100})
    #     f['t']['1.t1'].value = dt1
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': dt2,
    #             'd2': ['foo', 'abc'],
    #             'd3': [id1, 'xyz']
    #         },
    #         'l': [{'l1': 456, 'l2': 'baz'}, {'l1': 100, 'l2': '100'}],
    #         't': [789, {'t1': dt1, 't2': 'x'}],
    #     })
    #     self.assertDictEqual(f.self_value, {
    #         'd': {
    #             'd1': dt2,
    #             'd2': ['foo', 'abc'],
    #             'd3': [id1, 'xyz']
    #         },
    #         'l': [{'l1': 456, 'l2': 'baz'}, {'l1': 100, 'l2': '100'}],
    #         't': [789, {'t1': dt1, 't2': 'x'}],
    #     })
    #
    #     f.clear()
    #     self.assertDictEqual(f.value, {
    #         'd': {'d1': None, 'd2': [], 'd3': [None, None]},
    #         'l': [],
    #         't': [None, {'t1': None, 't2': None}],
    #     })
    #
    #     f.set_value({
    #         'd.d1': dt1.isoformat(),
    #         'd.d2.1': 'foo',
    #         'd.d2.2': 'bar',
    #         'd.d2.0': 'baz',
    #         'd.d3.1': 'xxx',
    #         'd.d3.0': str(id1),
    #         'l.1.l1': 456,
    #         'l.1.l2': 'baz2',
    #         'l.0.l1': 123,
    #         'l.0.l2': 'baz1',
    #         't.1.t1': dt2.isoformat(),
    #         't.0': 789
    #     })
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': dt1,
    #             'd2': ['baz', 'foo', 'bar'],
    #             'd3': [id1, 'xxx']
    #         },
    #         'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': 'baz2'}],
    #         't': [789, {'t1': dt2, 't2': None}],
    #     })
    #
    #     del f['d.d1']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': ['baz', 'foo', 'bar'],
    #             'd3': [id1, 'xxx']
    #         },
    #         'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': 'baz2'}],
    #         't': [789, {'t1': dt2, 't2': None}],
    #     })
    #
    #     del f['d']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': 'baz2'}],
    #         't': [789, {'t1': dt2, 't2': None}],
    #     })
    #
    #     del f['l.1.l2']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': None}],
    #         't': [789, {'t1': dt2, 't2': None}],
    #     })
    #
    #     del f['l.1']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [{'l1': 123, 'l2': 'baz1'}],
    #         't': [789, {'t1': dt2, 't2': None}],
    #     })
    #
    #     del f['l']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [],
    #         't': [789, {'t1': dt2, 't2': None}],
    #     })
    #
    #     del f['t.1']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [],
    #         't': [789, {'t1': None, 't2': None}],
    #     })
    #
    #     del f['t.0']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [],
    #         't': [None, {'t1': None, 't2': None}],
    #     })
    #
    #     del f['t.1.t1']
    #     self.assertDictEqual(f.value, {
    #         'd': {
    #             'd1': None,
    #             'd2': [],
    #             'd3': [None, None]
    #         },
    #         'l': [],
    #         't': [None, {'t1': None, 't2': None}],
    #     })
    #
    #     with self.assertRaises(IndexError):
    #         del f['t.2']
    #
    #     with self.assertRaises(KeyError):
    #         del f['foo']
    #
    # def test_dict_field_2(self):
    #     ch1 = ['a', 'b']
    #     ch2 = {1: 'x', 2: 'y'}
    #     data = {
    #         's': ChoiceField(ch1),
    #         'l': [ChoiceField(ch2)],
    #         't': (int, ChoiceField(ch2)),
    #     }
    #     f = make_field(data)
    #     f['s'].value = 'a'
    #     f['l'].append_value('x')
    #     f['t'].value = [123, 'y']
    #
    #     self.assertDictEqual(f.value, {'s': 'a', 'l': ['x'], 't': [123, 'y']})
    #     self.assertDictEqual(f.get_value(inner=True), {'s': 'a', 'l': [1], 't': [123, 2]})
    #     self.assertDictEqual(f.self_value, {'s': 'a', 'l': [1], 't': [123, 2]})
    #
    # def test_list_query_1(self):
    #     f = make_field([str])
    #     self.assertDictEqual(f.query, {})
    #     f.append_value('foo')
    #     self.assertDictEqual(f.query, {'$set': ['foo']})
    #     f.append_value('bar')
    #     self.assertDictEqual(f.query, {'$set': ['foo', 'bar']})
    #     del f[0]
    #     self.assertDictEqual(f.query, {'$set': ['bar']})
    #     f.dirty_clear()
    #     self.assertDictEqual(f.query, {})
    #     f[1].value = 'x'
    #     self.assertDictEqual(f.query, {'$set': ['bar', 'x']})
    #     f.pop()
    #     self.assertDictEqual(f.query, {'$set': ['bar']})
    #     f.dirty_clear()
    #
    #     del f[0]
    #     self.assertDictEqual(f.query, {'$set': []})
    #
    # def test_list_query_2(self):
    #     f = make_field([{'a': int, 'b': str}])
    #     self.assertDictEqual(f.query, {})
    #     f.append_value({'a': 1})
    #     self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}]})
    #     f.append_value({'b': 'foo'})
    #     self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}, {'a': None, 'b': 'foo'}]})
    #     f[0].value = {'a': 123, 'b': 'foo'}
    #     self.assertDictEqual(f.query, {'$set': [{'a': 123, 'b': 'foo'}, {'a': None, 'b': 'foo'}]})
    #
    # def test_list_query_3(self):
    #     f = make_field([{}])
    #     self.assertDictEqual(f.query, {})
    #     f.append_value({'a': 1})
    #     self.assertDictEqual(f.query, {'$set': [{'a': 1}]})
    #     f.append_value({'b': 'foo'})
    #     self.assertDictEqual(f.query, {'$set': [{'a': 1}, {'b': 'foo'}]})
    #     f[0].value = {'x': 123, 'y': 'foo'}
    #     self.assertDictEqual(f.query, {'$set': [{'a': 1, 'y': 'foo', 'x': 123}, {'b': 'foo'}]})
    #
    # def test_tuple_query_1(self):
    #     f = make_field((int, str))
    #     self.assertDictEqual(f.query, {})
    #     f[0].value = 123
    #     self.assertDictEqual(f.query, {'$set': [123, None]})
    #     f[1].value = 'foo'
    #     self.assertDictEqual(f.query, {'$set': [123, 'foo']})
    #     f.dirty_clear()
    #     self.assertDictEqual(f.query, {})
    #     del f[0]
    #     self.assertDictEqual(f.query, {'$set': [None, 'foo']})
    #
    # def test_tuple_query_2(self):
    #     f = make_field((str, [int]))
    #     self.assertDictEqual(f.query, {})
    #     f.set_value(['foo', [1]])
    #     self.assertDictEqual(f.query, {'$set': ['foo', [1]]})
    #     f['1.2'].value = 123
    #     self.assertDictEqual(f.query, {'$set': ['foo', [1, None, 123]]})
    #
    # def test_tuple_query_3(self):
    #     f = make_field((int, {'a': int}))
    #     self.assertDictEqual(f.query, {})
    #     f['1.a'].value = 123
    #     self.assertDictEqual(f.query, {'$set': [None, {'a': 123}]})
    #     f[0].value = 123
    #     self.assertDictEqual(f.query, {'$set': [123, {'a': 123}]})

    def test_dict_query_1(self):
        data = {
            'd1': int,
            'd2': [str],
            'd3': (int, str),
            'd4': {
                'dd1': int,
                'dd2': [str],
                'dd3': (int, str)
            },
            'd5': {}
        }
        f = make_field(data)
        self.assertDictEqual(f.query, {})
        f['d1'].value = 123
        print('---1', f.dirty_fields)

        self.assertDictEqual(f.query, {'$set': {'d1': 123}})
        print('---1', f.dirty_fields)
        f['d2.0'].value = 123
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['123']
        }})
        f['d2'][1].value = 'foo'
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['123', 'foo']
        }})
        f['d2'].append_value('bar')
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['123', 'foo', 'bar']
        }})
        f['d2'].set_value(['z', 123])
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar']
        }})
        f['d3.0'].value = 123
        f['d3.1'].value = 123
        print('---3', f.dirty_fields)
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [123, '123']
        }})
        f['d3'].value = [None, 'foo']
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo']
        }})
        f['d4.dd1'].value = 456
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456
        }})
        f['d4.dd2'].append_value('bar')
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456,
            'd4.dd2': ['bar']
        }})
        f['d4.dd2.2'].value = 'x'
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456,
            'd4.dd2': ['bar', None, 'x']
        }})
        f['d4.dd3.1'].value = 456
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456,
            'd4.dd2': ['bar', None, 'x'],
            'd4.dd3': [None, '456']
        }})
        print('---4', f.dirty_fields)
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        f.set_value({
            'd1': 456,
            'd2': ['foo', 'bar'],
            'd3': [123, 'x'],
            'd4': {
                'dd1': 123,
                'dd2': ['a', 'b', 'c'],
                'dd3': [None, None]
            }
        })
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar', 'bar'],
            'd3': [123, 'x'],
            'd4.dd1': 123,
            'd4.dd2': ['a', 'b', 'c'],
            'd4.dd3': [None, None]
        }})
        f['d5'].value = {'a': 1, 'b': 'foo'}
        print('---5', f.dirty_fields)
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar', 'bar'],
            'd3': [123, 'x'],
            'd4.dd1': 123,
            'd4.dd2': ['a', 'b', 'c'],
            'd4.dd3': [None, None],
            'd5': {'a': 1, 'b': 'foo'}
        }})

        f['d4.dd3.1'].value = 123
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar', 'bar'],
            'd3': [123, 'x'],
            'd4.dd1': 123,
            'd4.dd2': ['a', 'b', 'c'],
            'd4.dd3': [None, '123'],
            'd5': {'a': 1, 'b': 'foo'}
        }})

    # def test_dict_query_2(self):
    #     f = make_field({})
    #     self.assertDictEqual(f.query, {})
    #     f.set_value({
    #         'd1': 456,
    #         'd2': ['foo', 'bar'],
    #         'd3': [123],
    #         'd4': {
    #             'dd1': 123,
    #             'dd2': ['a', 'b', 'c'],
    #         }
    #     })
    #     self.assertDictEqual(f.query, {'$set': {
    #         'd1': 456,
    #         'd2': ['foo', 'bar'],
    #         'd3': [123],
    #         'd4': {
    #             'dd1': 123,
    #             'dd2': ['a', 'b', 'c'],
    #         }
    #     }})



from .documents import Document, Documents
from .models import Model
# from .manage import ModelManager
from .fields import ChoiceField as Choice
from .geo_fileds import (
    GEOPoint,
    GEOLine,
    GEOPolygon,
    GEOMultiPoint,
    GEOMultiLine,
    GEOMultiPolygon
)
import mokito.errors as errors

__all__ = [
    "Document",
    "Documents",
    "Model",
    # "ModelManager",
    "errors",
    "Choice",
    "GEOPoint",
    "GEOLine",
    "GEOPolygon",
    "GEOMultiPoint",
    "GEOMultiLine",
    "GEOMultiPolygon"
]

version = "0.3.9"

# TODO: add indexes
# TODO: add forms
# TODO: add GraphQL
# TODO: add test .reload



import copy

from .fields import Field
from .tools import make_field, SEPARATOR
from .errors import MokitoDBREFError


class CollectionField(Field):
    def __init__(self, _parent=None, **kwargs):
        super().__init__(_parent=_parent, **kwargs)
        self._val = {}
        self._dirty_fields = {}

    @property
    def dirty_fields(self):
        return list(self._dirty_fields.keys())

    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def _test_item(self, key):
        raise NotImplementedError

    def _sep_key(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        return k1, k2


class ArrayField(CollectionField):
    def __len__(self):
        return max(map(int, self._val.keys())) + 1 if self._val else 0

    def __iter__(self):
        for i in range(len(self)):
            yield self.__getitem__(i)

    def _sep_key(self, key):
        k1, k2 = super()._sep_key(key)
        try:
            ik1 = int(k1)
        except ValueError:
            raise TypeError("%s indices must be integers" % self.__class__.__name__)
        if ik1 < 0:
            ik1 += len(self)
            k1 = str(ik1)
        return k1, k2

    @property
    def dirty_fields(self):
        return list(map(int, self._dirty_fields.keys()))

    def dirty_clear(self):
        from .documents import Document
        self.dirty = False
        for i in self._val.values():
            if not isinstance(i, Document):
                i.dirty_clear()
        self._dirty_fields = {}

    def get_dirty(self):
        from .documents import Document
        if not self._dirty:
            for i in self._val.values():
                if not isinstance(i, Document) and i.dirty:
                    self.dirty = True
                    break
        return self._dirty

    dirty = property(get_dirty, CollectionField.set_dirty)

    def set_value(self, value, **kwargs):
        from .documents import Document, Model

        if isinstance(value, (list, tuple)):
            value = {str(k): v for k, v in enumerate(value)}

        elif value is None:
            self.clear()
            return

        if not isinstance(value, dict):
            raise TypeError()

        value = self.validate(value, **kwargs)

        for key, v in value.items():
            k1, k2 = self._sep_key(key)
            item = self._test_item(k1)

            if k2:
                item.set_value({k2: v}, **kwargs)
            else:
                if isinstance(item, Document):
                    if v is None:
                        del self[k1]
                    elif isinstance(v, Document):
                        self._val[k1] = v
                        if item.id != v.id:
                            self._dirty = True
                    else:
                        item.set_value(v, **kwargs)

                elif isinstance(v, Model):
                    item.set_value(v.value, **kwargs)

                else:
                    item.set_value(v, **kwargs)

    def get_value(self, **kwargs):
        return [self._val[i].get_value(**kwargs) if i in self._val else None for i in map(str, range(len(self)))]

    value = property(get_value, set_value)

    def update_values(self, *args, **kwargs):
        for k, v in enumerate(args):
            self[k].value = v

        for k, v in kwargs.items():
            self[k].value = v

    @property
    def self_value(self):
        return [self._val[i].self_value if i in self._val else None for i in map(str, range(len(self)))]

    def make_query(self, short=False):
        from .documents import Document

        if not self.dirty:
            return {}

        ret = {"$set": []}
        for k in range(len(self)):
            k = str(k)
            v = self._val.get(k)

            if v is None:
                ret["$set"].append(None)

            elif isinstance(v, Document):
                if v.dirty or v.make_query(True):
                    if v.id is None:
                        raise MokitoDBREFError(v)
                ret["$set"].append(v.dbref)

            else:
                ret["$set"].append(v.get_value(inner=True))

            if short and ret['$set']:
                return True
        return ret

    query = property(make_query)


class ListField(ArrayField):
    def __init__(self, rules, _parent=None, **kwargs):
        super().__init__(_parent=_parent, **kwargs)
        if not rules:
            rules = [None]
        self._default = False
        _rules = [make_field(i, _parent=self) for i in rules]
        self._rules = copy.deepcopy(_rules[0])
        for k, v in enumerate(_rules):
            if v._default:
                self._default = True
                v._parent = self
                self._val[str(k)] = v

    def _test_item(self, key):
        key = str(key)
        if key not in self._val:
            self.dirty = True
            item = copy.deepcopy(self._rules)
            item._parent = self
            self._val[key] = item
        return self._val[key]

    def __getitem__(self, key):
        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        try:
            return item[k2] if k2 else item
        except TypeError:
            raise IndexError('Out of range')

    def __setitem__(self, key, value):
        from .fields import AnyField

        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        if k2:
            item[k2] = value
        else:
            if value is None:
                del self[k1]
            elif isinstance(self._rules, AnyField) or value.__class__ == self._rules.__class__:
                self._val[k1] = value
            else:
                raise TypeError('%s is not %s' % (value.__class__.__name__, self._rules.__class__.__name__))

    def __delitem__(self, key):
        k1, k2 = self._sep_key(key)
        if not k2 and k1 in self._val:
            self.dirty = True
        item = self._test_item(k1)
        if k2:
            del item[k2]

        else:
            del self._val[k1]
            keys = [int(i) for i in self._val.keys()]
            keys.sort()
            ik1 = int(k1)
            for i in keys:
                if i > ik1:
                    item = self._val.pop(str(i))
                    if item.value is not None:
                        self._val[str(i - 1)] = item

    def clear(self):
        self._val = {}
        self.dirty = True

    def append(self, value):
        self[len(self)] = value

    def append_value(self, value, **kwargs):
        item = self._test_item(str(len(self)))
        value = self.validate(value, **kwargs)
        item.set_value(value, **kwargs)

    def pop(self, key=-1):
        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        if k2:
            return item.pop(k2)
        del self[k1]
        return item


class TupleField(ArrayField):
    def __init__(self, rules, _parent=None, **kwargs):
        super().__init__(_parent=_parent, **kwargs)
        self._default = False
        for k, v in enumerate(rules):
            item = make_field(v, _parent=self)
            if item._default:
                self._default = True
            self._val[str(k)] = item

    def _test_item(self, key):
        try:
            return self._val[key]
        except KeyError:
            raise IndexError("%s out of range" % key)

    def __getitem__(self, key):
        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        try:
            return item[k2] if k2 else item
        except TypeError:
            raise IndexError('Out of range')

    def __setitem__(self, key, value):
        from .fields import AnyField

        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        if k2:
            item[k2] = value
        else:
            if value is None:
                del self[k1]
            elif isinstance(item, AnyField) or value.__class__ == item.__class__:
                self._val[k1] = value
            else:
                raise TypeError('%s not %s' % (value.__class__.__name__, item.__class__.__name__))

    def __delitem__(self, key):
        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        if k2:
            del item[k2]
        else:
            item.set_value(None)
        self.dirty = True

    def clear(self):
        for k, v in self._val.items():
            if isinstance(v, (ArrayField, DictField)):
                v.clear()
            else:
                self[k] = None

        self.dirty = True


class DictField(CollectionField):
    def __init__(self, rules, _parent=None, **kwargs):
        super().__init__(_parent=_parent, **kwargs)
        # TODO: убрать безрулевые, т.е. {}
        self._default = False
        self._rules = bool(rules)
        if rules:
            for k, v in rules.items():
                item = make_field(v, _parent=self)
                if item._default:
                    self._default = True
                self._val[k] = item

    def __len__(self):
        return len(self._val)

    def _test_item(self, key):
        from .fields import AnyField

        if not self._rules and key not in self._val:
            self._val[key] = AnyField(_parent=self)
        return self._val[key]

    def __getitem__(self, key):
        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        return item[k2] if k2 else item

    def __setitem__(self, key, value):
        from .fields import AnyField

        k1, k2 = self._sep_key(key)
        item = self._test_item(k1)
        if k2:
            item[k2] = value

        else:
            if value is None:
                self._dirty_fields[k1] = True
                self[k1].set_value(None)

            elif not self._rules or isinstance(item, AnyField) or value.__class__ == item.__class__:
                self._dirty_fields[k1] = True
                self._val[k1] = value

            else:
                raise TypeError('%s not %s' % (value.__class__.__name__, item.__class__.__name__))

    def __delitem__(self, key):
        k1, k2 = self._sep_key(key)
        if k2:
            del self._val[k1][k2]
        else:
            if not self._rules:
                del self._val[k1]
            else:
                self._val[k1].clear()
            self._dirty_fields[k1] = False

    def get_value(self, **kwargs):
        return {k: v.get_value(**kwargs) for k, v in self._val.items()}

    def set_value(self, value, **kwargs):
        from .documents import Document, Model

        if value is None:
            self.clear()
            return

        if not isinstance(value, dict):
            raise TypeError()

        value = self.validate(value, **kwargs)

        for key, v in value.items():
            k1, k2 = self._sep_key(key)
            try:
                item = self._test_item(k1)
            except KeyError:
                if kwargs.get('safe'):
                    continue
                else:
                    raise

            if k2:
                item.set_value({k2: v}, **kwargs)
            else:
                if isinstance(item, Document):
                    if v is None:
                        del self[k1]
                    elif isinstance(v, Document):
                        self._val[k1] = v
                        if item.id != v.id:
                            self._dirty_fields[k1] = True
                    else:
                        item.set_value(v, **kwargs)

                elif isinstance(v, Model):
                    item.set_value(v.value, **kwargs)

                else:
                    item.set_value(v, **kwargs)

    value = property(get_value, set_value)

    def update_values(self, **kwargs):
        for k, v in kwargs.items():
            self[k].value = v

    @property
    def self_value(self):
        return {k: v.self_value for k, v in self._val.items()}

    def dirty_clear(self):
        from .documents import Document
        for i in self._val.values():
            if not isinstance(i, Document):
                i.dirty_clear()
        self._dirty_fields = {}

    def items(self):
        return self._val.items()

    @property
    def dirty(self):
        from .documents import Document
        if self._dirty_fields:
            return True
        for k, v in self._val.items():
            if not isinstance(v, Document) and v.dirty:
                self._dirty_fields[k] = True
                # print('X5', self._dirty_fields)
                return True
        return False

    def clear(self):
        self._dirty_fields = {i: False for i in self._val.keys()}
        if self._rules:
            for i in self._val.values():
                i.clear()
        else:
            self._val = {}

    def make_query(self, short=False):
        from .documents import Document
        # from .models import Model
        ret = {"$set": {}, "$unset": {}}

        if not self._rules:
            ret['$set'] = self.get_value(inner=True)

        else:
            for k, v in self._val.items():
                if isinstance(v, Document):
                    _d = self._dirty_fields.get(k, None)
                    if _d is False:
                        ret["$unset"][k] = ''
                    elif (_d is True) or v.ref_changed:
                        if v.id is None:
                            raise MokitoDBREFError(v)
                        ret["$set"][k] = v.dbref

                elif k in self._dirty_fields or v.dirty:
                    # print('X7', self._dirty_fields)
                    if isinstance(v, ArrayField):
                        _q = v.query
                        if _q:
                            ret["$set"][k] = _q["$set"]

                    # elif isinstance(v, Model):
                    #     _q = v.query
                    #     for j in ["$set", "$unset"]:
                    #         if j in _q:
                    #             for k1, v1 in _q[j].items():
                    #                 ret[j]["%s.%s" % (k, k1)] = v1

                    elif isinstance(v, DictField):
                        _q = v.query
                        for j in ["$set", "$unset"]:
                            if j in _q:
                                if v._rules:
                                    for k1, v1 in _q[j].items():
                                        ret[j]["%s.%s" % (k, k1)] = v1
                                else:
                                    ret[j][k] = _q[j]
                    else:
                        ret['$set'][k] = v.value

                    if short and (ret['$set'] or ret['$unset']):
                        return True

        if not ret["$set"]:
            del ret["$set"]
        if not ret["$unset"]:
            del ret["$unset"]
        return ret

    query = property(make_query)



import copy
from collections import UserList

from bson import DBRef, ObjectId
from pymongo import ASCENDING, DESCENDING

from .models import Model
from .errors import MokitoORMError
from .tools import SEPARATOR


class Documents(UserList):
    def __repr__(self):
        return '<%s: [%s]>' % (self.__class__.__name__, ', '.join(map(str, self.data)))

    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        ret = self.data[int(k1)]
        return ret[k2] if k2 else ret

    # def dirty_clear(self):
    #     map(lambda i: i.dirty_clear(), self.data)

    async def reload(self, *fields, cash=None):
        if cash is None:
            cash = {}
        for i in self.data:
            await i.reload(*fields, cash=cash)

    async def remove(self, safe=True):
        if self.data:
            ids = [i.id for i in self.data]
            await self.data[0].get_collection().delete_many({"_id": {"$in": ids}})
            for i in self.data:
                i.id = None

    async def save(self):
        for i in self.data:
            await i.save()

    # def filter(self, fn):
    #     return Documents(filter(fn, self.data))

    def get_value(self, **kwargs):
        return [i.get_value(**kwargs) for i in self.data]

    value = property(get_value)

    def as_json(self, *args, **kwargs):
        return [i.as_json(*args, **kwargs) for i in self.data]

    def as_dict(self, *args, **kwargs):
        return {str(i.id): i.as_json(*args, **kwargs) for i in self.data}


class Result(object):
    def __init__(self, cls, db_cursor):
        self.cls = cls
        self.db_cursor = db_cursor

    def __aiter__(self):
        return self

    async def __anext__(self):
        if await self.db_cursor.fetch_next:
            data = self.db_cursor.next_object()
            return self.cls.mk(data)
        raise StopAsyncIteration

    async def all(self):
        return Documents([self.cls.mk(data) async for data in self.db_cursor])


class Document(Model):
    __database__ = None
    __collection__ = None
    sorting = []

    def __init__(self, data=None, **kwargs):
        self.loaded = False
        self.ref_changed = False
        self._id = None
        if data:
            _id = data.pop('_id', None)
            if _id:
                self._id = ObjectId(_id)
        super().__init__(data, **kwargs)

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self._id)

    def __getitem__(self, key):
        if key == 'id':
            return self.id
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        if key == 'id':
            self.id = value
        else:
            super().__setitem__(key, value)

    @property
    def self_value(self):
        return self.get_dbref() if self._id else None

    def _set_parent_dirty(self):
        self._dirty = True
        self.ref_changed = True
        if self._parent is not None:
            self._parent._dirty = True

    def get_id(self):
        return self._id

    def set_id(self, value):
        if value is not None:
            value = ObjectId(value)
        if self._id != value:
            self._id = value
            self._set_parent_dirty()

    id = property(get_id, set_id)

    def get_dbref(self):
        return DBRef(self.__collection__, self._id)

    def set_dbref(self, dbref):
        if self.__collection__ != dbref.collection:
            self.__collection__ = dbref.collection
            self._set_parent_dirty()
        self.set_id(dbref.id)

    dbref = property(get_dbref, set_dbref)

    @classmethod
    def mk(cls, data):
        self = cls(data, inner=True, safe=True)
        self.dirty_clear()
        self.loaded = True
        return self

    def clear(self):
        super().clear()
        self._id = None

    def dirty_clear(self):
        super().dirty_clear()
        self.ref_changed = False

    def get_value(self, **kwargs):
        data = super().get_value(**kwargs)
        data['_id'] = self._id
        return data

    def set_value(self, value, **kwargs):
        if isinstance(value, DBRef):
            self.dbref = value
        else:
            if isinstance(value, dict):
                value = copy.deepcopy(value)
                _id = value.pop('_id', None)
                if _id is not None:
                    self._id = ObjectId(_id)
            super().set_value(value, **kwargs)

    value = property(get_value, set_value)

    @classmethod
    def get_collection(cls, database=None, collection=None):
        return (database or cls.__database__)[collection or cls.__collection__]

    @staticmethod
    def norm_spec(spec):
        if spec is None:
            return {}
        if isinstance(spec, bytes):
            return {"_id": ObjectId(spec.decode('utf-8'))}
        if isinstance(spec, str):
            return {"_id": ObjectId(spec)}
        if isinstance(spec, ObjectId):
            return {"_id": spec}
        return spec

    @classmethod
    def norm_sort(cls, sort):
        _sort = sort or cls.sorting
        if not _sort:
            return
        if not isinstance(_sort, list):
            _sort = [_sort]
        return [(i[1:], DESCENDING) if i[0] == '-' else (i, ASCENDING) for i in _sort]

    # @classmethod
    # @coroutine
    # def find_one_raw(cls, spec_or_id, *fields):
    #     data = yield cls.get_collection().find_one(spec_or_id, fields=fields)
    #     raise Return(data)

    @classmethod
    async def find_one(cls, spec_or_id):
        spec = cls.norm_spec(spec_or_id)
        data = await cls.get_collection().find_one(spec, list(cls.scheme.keys()))
        if data:
            return cls.mk(data)

    # @classmethod
    # @coroutine
    # def find_raw(cls, spec, *fields, **kwargs):
    #     kw = {
    #         'fields': fields,
    #         'skip': kwargs.get('skip', 0),
    #         'limit': kwargs.get('limit', 0),
    #         'sort': kwargs.get('sort'),
    #         'hint': kwargs.get('hint')
    #     }
    #     data = yield cls.get_collection().find(spec, **kw)
    #     raise Return(data)

    @classmethod
    def find(cls, spec_or_id=None, skip=0, limit=0, sort=None):
        if skip is None:
            skip = 0
        if limit is None:
            limit = 0
        spec = cls.norm_spec(spec_or_id)
        sort = cls.norm_sort(sort)
        cursor = cls.get_collection().find(spec, list(cls.scheme.keys()), skip=skip, limit=limit, sort=sort)
        return Result(cls, cursor)

    @classmethod
    async def count(cls, spec=None):
        spec = cls.norm_spec(spec)
        return await cls.get_collection().count(spec)

    @classmethod
    async def distinct(cls, key, spec=None):
        spec = cls.norm_spec(spec)
        return await cls.get_collection().distinct(key, spec)

    async def reload(self, *fields, cash=None):
        """
        Read the object again.
        """
        if cash is None:
            cash = {}
        if fields:
            return any([await self[i].reload(cash=cash) for i in fields])

        elif self._id:
            k = self.dbref
            if k in cash:
                data = cash[k]
            else:
                data = await self.get_collection().find_one({'_id': self._id}, list(self.scheme.keys()))
                cash[k] = data

            if data:
                self.clear()
                self.set_value(data, inner=True, safe=True)
                self.dirty_clear()
                self.loaded = True
                return True

    async def save(self):
        if self._id is None:
            res = await self.get_collection().insert_one(super().self_value)
            self._id = res.inserted_id
            self.dirty_clear()
            return True

        elif self.dirty:
            await self.get_collection().update_one({'_id': self._id}, self.query)
            self.dirty_clear()
            return True

        return False

    async def remove(self):
        if self._id is not None:
            res = await self.get_collection().delete_one({'_id': self._id})
            self._id = None
            return res.deleted_count
        return 0

    @classmethod
    async def delete(cls, spec=None):
        spec = cls.norm_spec(spec)
        res = await cls.get_collection().delete_many(spec)
        return res.deleted_count



from .fields import FloatField, StringField
from .collections import TupleField, ListField, DictField


class _ConstString(StringField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        super().__init__(_default, _parent, **kwargs)
        self._val = self.validate(_default)

    def set_value(self, value, **kwargs):
        pass

    value = property(StringField.get_value, set_value)


class _Longitude(FloatField):
    def validate(self, value, **kwargs):
        if value is not None:
            value = float(value)
            if value < -180.0 or value > 180.0:
                raise ValueError()
        return value


class _Latitude(FloatField):
    def validate(self, value, **kwargs):
        if value is not None:
            value = float(value)
            if value < -90.0 or value > 90.0:
                raise ValueError()
        return value


class _Point1(TupleField):
    """
    [lng, lat]
    """

    def __init__(self, _default=None, _parent=None, **kwargs):
        if _default is None:
            _default = [None, None]

        rules = [_Longitude(_default=_default[0]),
                 _Latitude(_default=_default[1])]
        super().__init__(rules, _parent=_parent, **kwargs)

    def get_long(self):
        return self[0].value

    def set_long(self, value):
        self[0].value = value

    longitude = property(get_long, set_long)

    def get_lat(self):
        return self[1].value

    def set_lat(self, value):
        self[1].value = value

    latitude = property(get_lat, set_lat)


class _Point2(ListField):
    """
    [
        [lng, lat], ..., [lngN, latN]
    ]
    """

    def __init__(self, _default=None, _parent=None, **kwargs):
        if _default is None:
            rules = [_Point1()]
        else:
            rules = [_Point1(i) for i in _default]
        super().__init__(rules, _parent=_parent, **kwargs)

    def get_value(self, **kwargs):
        return [self._test_item(i).get_value(**kwargs) for i in map(str, range(len(self)))]

    value = property(get_value, ListField.set_value)


class _Point3(ListField):
    """
    [
        [ [lng1, lat1], ..., [lng1N, lat1N] ],
        ...
        [ [lngP, latP], ..., [lngPN, latPN] ]
    ]
    """

    def __init__(self, _default=None, _parent=None, **kwargs):
        if _default is None:
            rules = [_Point2()]
        else:
            rules = [_Point2(i) for i in _default]
        super().__init__(rules, _parent=_parent, **kwargs)

    def get_value(self, **kwargs):
        return [self._test_item(i).get_value(**kwargs) for i in map(str, range(len(self)))]

    value = property(get_value, ListField.set_value)


class _Point4(_Point3):
    """
    [
        [ [lng1, lat1], ..., [lng1N, lat1N] ],
        ...
        [ [lngP, latP], ..., [lngPN, latPN] ]
    ]
    """

    def get_value(self, **kwargs):
        ret = super().get_value(**kwargs)
        if ret:
            for k, v in enumerate(ret):
                if v and v[0] != v[-1]:
                    ret[k].append(v[0])
        return ret

    value = property(get_value, _Point3.set_value)


class _Point5(ListField):
    """
    [
        [
            [ [lng1, lat1], ..., [lng1N, lat1N] ],
            ...
            [ [lngP, latP], ..., [lngPN, latPN] ]
        ],
        ...
        [
            [ [lng1, lat1], ..., [lng1N, lat1N] ],
            ...
            [ [lngP, latP], ..., [lngPN, latPN] ]
        ]
    ]
    """

    def __init__(self, _default=None, _parent=None, **kwargs):
        if _default is None:
            rules = [_Point4()]
        else:
            rules = [_Point4(i) for i in _default]
        super().__init__(rules, _parent=_parent, **kwargs)

    def get_value(self, **kwargs):
        return [self._test_item(i).get_value(**kwargs) for i in map(str, range(len(self)))]

    value = property(get_value, ListField.set_value)


class GEOField(DictField):
    def make_query(self, short=False):
        if self.dirty:
            return True if short else {'$set': self.value}
        return {}

    query = property(make_query)


class GEOPointField(GEOField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        rules = {
            'type': _ConstString('Point'),
            'coordinates': _Point1(_default)
        }
        super().__init__(rules, _parent=_parent, **kwargs)

GEOPoint = GEOPointField


class GEOLineField(GEOField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        rules = {
            'type': _ConstString('LineString'),
            'coordinates': _Point2(_default)
        }
        super().__init__(rules, _parent=_parent, **kwargs)

GEOLine = GEOLineField


class GEOPolygonField(GEOField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        rules = {
            'type': _ConstString('Polygon'),
            'coordinates': _Point4(_default)
        }
        super().__init__(rules, _parent=_parent, **kwargs)

GEOPolygon = GEOPolygonField


class GEOMultiPointField(GEOField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        rules = {
            'type': _ConstString('MultiPoint'),
            'coordinates': _Point2(_default)
        }
        super().__init__(rules, _parent=_parent, **kwargs)

GEOMultiPoint = GEOMultiPointField


class GEOMultiLineField(GEOField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        rules = {
            'type': _ConstString('MultiLineString'),
            'coordinates': _Point3(_default)
        }
        super().__init__(rules, _parent=_parent, **kwargs)

GEOMultiLine = GEOMultiLineField


class GEOMultiPolygonField(GEOField):
    def __init__(self, _default=None, _parent=None, **kwargs):
        rules = {
            'type': _ConstString('MultiPolygon'),
            'coordinates': _Point5(_default)
        }
        super().__init__(rules, _parent=_parent, **kwargs)

GEOMultiPolygon = GEOMultiPolygonField



import datetime

from bson import ObjectId, DBRef
import motor.motor_tornado

import mokito

TEST_DB_NAME = 'test_mokito'
TEST_DB_URI = 'mongodb://127.0.0.1:27017'
TEST_COLLECTION1 = 'tst1'
TEST_COLLECTION2 = 'tst2'
TEST_COLLECTION3 = 'tst3'

db = motor.motor_tornado.MotorClient(TEST_DB_URI)[TEST_DB_NAME]


dt1 = datetime.datetime(2016, 1, 2, 3, 4, 5)
dt2 = datetime.datetime(2016, 2, 3, 4, 5, 6)
dt3 = datetime.datetime(2016, 3, 4, 5, 6, 7)
dt4 = datetime.datetime(2016, 4, 5, 6, 7, 8)
dt5 = datetime.datetime(2016, 5, 6, 7, 8, 9)
dt6 = datetime.datetime(2016, 6, 7, 8, 9, 10)
dt7 = datetime.datetime(2016, 7, 8, 9, 10, 11)
dt8 = datetime.datetime(2016, 8, 9, 10, 11, 12)
dt9 = datetime.datetime(2016, 9, 10, 11, 12, 13)
dt10 = datetime.datetime(2016, 10, 11, 12, 13, 14)
dt11 = datetime.datetime(2016, 11, 12, 13, 14, 15)
dt12 = datetime.datetime(2016, 12, 13, 14, 15, 16)
xt1 = datetime.datetime(2015, 1, 2, 3, 4, 5)
xt2 = datetime.datetime(2015, 2, 3, 4, 5, 6)
xt3 = datetime.datetime(2015, 3, 4, 5, 6, 7)


col1_id1 = ObjectId()
col1_dbref1 = DBRef(TEST_COLLECTION1, col1_id1)
col1_data1 = {
    '_id': col1_id1,
    'x1': 0.5,
    'x2': [dt1, dt2, dt3],
    'x3': [123, 'z1'],
    'x4': {'a': 1, 'b': 2}
}

col1_id2 = ObjectId()
col1_dbref2 = DBRef(TEST_COLLECTION1, col1_id2)
col1_data2 = {
    '_id': col1_id2,
    'x1': 1.5,
    'x2': [dt4, dt5, dt6],
    'x3': [45601, 'z2'],
    'x4': {'a': 3, 'b': 4},
    'foo': 'bar'
}


col2_id1 = ObjectId()
col2_data1 = {
    '_id': col2_id1,
    'f1': (123, 'foo2'),
    'f2': {'a': 5, 'b': 6},
    'm1': {
        'x1': .1,
        'x2': [dt7, dt8],
        'x3': [7, 'z3'],
        'x4': {'a': 8, 'b': 8}
    },
    'm2': [{
            'x1': .2,
            'x2': [dt9],
            'x3': [8, 'z1'],
            'x4': {'a': 9, 'b': 9}
        }, {
            'x1': .3,
            'x2': [dt10],
            'x3': [9, 'z2'],
            'x4': {'a': 10, 'b': 10}
    }],
    'd1': col1_dbref1,
    'd2': [col1_dbref1, col1_dbref2]
}


col2_id2 = ObjectId()
col2_data2 = {
    '_id': col2_id2,
    'f1': [45602, 'foo3'],
    'f2': {'a': 11, 'b': 12},
    'm1': {
        'x1': .2,
        'x2': [dt11, dt12],
        'x3': [7, 'z3'],
        'x4': {'a': 8, 'b': 8}
    },
    'm2': [{
            'x1': .2,
            'x2': [xt1],
            'x3': [8, 'z1'],
            'x4': {'a': 9, 'b': 9}
        }, {
            'x1': .3,
            'x2': [xt2, xt3],
            'x3': [9, 'z2'],
            'x4': {'a': 10, 'b': 10}
    }],
    'd1': col1_dbref2,
    'd2': [col1_dbref2]
}


col3_id1 = ObjectId()
col3_data1 = {
    '_id': col3_id1,
    'foo': 1,
    'bar': 'zzz'
}

col3_id2 = ObjectId()
col3_data2 = {
    '_id': col3_id2,
    'foo': 2,
    'bar': 'aaa'
}


class Document0(mokito.Document):
    __database__ = db


ch1 = {'z1': 1, 'z2': 2, 'z3': 3}


class Model0(mokito.Model):
    scheme = {
        'x1': float,
        'x2': [datetime.datetime],
        'x3': (int, mokito.Choice(ch1)),
        'x4': {'a': int, 'b': int}
    }


class Model1(Model0):
    @property
    def prop1(self):
        return self['x1'].value + 2


class Document1(Document0):
    __collection__ = TEST_COLLECTION1
    scheme = Model1

    prop1 = Model1.prop1

    @property
    def prop2(self):
        return self._val['x1'].value * 2


class Model2(mokito.Model):
    scheme = {
        'f1': (int, str),
        'f2': {'a': int, 'b': int},
        'm1': Model1,
        'm2': [Model1],
        'd1': Document1,
        'd2': [Document1]
    }


class Document2(Document0):
    __collection__ = TEST_COLLECTION2
    scheme = Model2


class Document3(Document0):
    __collection__ = TEST_COLLECTION3
    scheme = {
        'foo': int,
        'bar': str,
    }


class Document4(Document3):
    sorting = ['-foo']

# TODO: add test update_values



from bson import ObjectId

from .manage import ModelManager
from .collections import DictField


class ModelMeta(type):
    def __new__(mcs, name, bases, attr):
        scheme = attr.get('scheme')
        if scheme is not None and not isinstance(scheme, dict):
            attr['scheme'] = getattr(scheme, 'scheme')

        _cls = type.__new__(mcs, name, bases, attr)
        if name != 'Model' and name != 'Document':
            ModelManager.add(_cls)
        return _cls


class Model(DictField, metaclass=ModelMeta):
    scheme = {}

    def __init__(self, data=None, **kwargs):
        super().__init__(self.scheme, **kwargs)
        if data:
            self.set_value(data, **kwargs)

    def as_json(self, *args, **kwargs):
        def _set_v():
            if hasattr(self, k):
                v = getattr(self, k, None)
                if hasattr(v, '__call__'):
                    v = v()
            else:
                v = self[k]
                if hasattr(v, 'get_value'):
                    v = v.get_value(_format='json')

            if isinstance(v, ObjectId):
                v = str(v)
            res[alias] = v

        res = {}
        for k in args:
            alias = k
            _set_v()
        for alias, k in kwargs.items():
            _set_v()
        return res

    # def get(self, key=None, **kwargs):
    #     if isinstance(key, (list, tuple)):
    #         data = {}
    #         fields = []
    #         for i in key:
    #             if hasattr(self, i):
    #                 data[i] = getattr(self, i)
    #             else:
    #                 fields.append(i)
    #         return dict(super(Model, self).get(fields, **kwargs), **data)
    #
    #     return super(Model, self).get(key, **kwargs)



# import unittest
#
# from mokito.fields import AnyField
# from mokito.tools import make_field
# from mokito.collections import ListField, TupleField
#
#
# class TestNoneFields(unittest.TestCase):
#     def test_list_field_1(self):
#         f = make_field([None])
#         self.assertIsInstance(f, ListField)
#         self.assertIsInstance(f._rules, AnyField)
#         f[1].value = 123
#         f.append_value('foo')
#         self.assertIsNone(f.pop(0).value)
#         self.assertListEqual(f.value, [123, 'foo'])
#         f[1].value = None
#         self.assertListEqual(f.value, [123, None])
#         del f[1]
#         self.assertListEqual(f.value, [123])
#
#         f.append_value(('bar', 2))
#         self.assertListEqual(f.value, [123, ['bar', 2]])
#         self.assertEqual(f[1][1].value, 2)
#         self.assertEqual(f['1.1'].value, 2)
#
#     def test_tuple_field_1(self):
#         f = make_field((None, str))
#         self.assertIsInstance(f, TupleField)
#         self.assertListEqual(f.value, [None, None])
#         f[0].value = 123
#         f[1].value = 123
#         self.assertListEqual(f.value, [123, '123'])
#         f[0].value = 'foo'
#         self.assertListEqual(f.value, ['foo', '123'])
#         del f[0]
#         self.assertListEqual(f.value, [None, '123'])
#
#     def test_tuple_field_2(self):
#         f = make_field((str, [None]))
#         self.assertListEqual(f.value, [None, []])
#         f.set_value(['foo', [1, 'bar']])
#         self.assertListEqual(f.value, ['foo', [1, 'bar']])
#         f[1].append_value(123.4)
#         self.assertListEqual(f.value, ['foo', [1, 'bar', 123.4]])
#         del f['1.1']
#         self.assertListEqual(f.value, ['foo', [1, 123.4]])
#         f[1].value = None
#         self.assertListEqual(f.value, ['foo', []])
#
#     def test_dict_field_1(self):
#         f = make_field({
#             'a': int,
#             'b': None
#         })
#         self.assertDictEqual(f.value, {'a': None, 'b': None})
#
#         f.set_value({'a': 123, 'b': 456})
#         self.assertDictEqual(f.value, {'a': 123, 'b': 456})
#
#         f.set_value({'a': 1, 'b': [1, 2, 3]})
#         self.assertDictEqual(f.value, {'a': 1, 'b': [1, 2, 3]})
#
#         f.clear()
#         self.assertDictEqual(f.value, {'a': None, 'b': None})
#
#     def test_list_query_1(self):
#         f = make_field([None])
#         self.assertDictEqual(f.query, {})
#         f.append_value('foo')
#         self.assertDictEqual(f.query, {'$set': ['foo']})
#         f.append_value(123)
#         self.assertDictEqual(f.query, {'$set': ['foo', 123]})
#         del f[0]
#         self.assertDictEqual(f.query, {'$set': [123]})
#         f.dirty_clear()
#         self.assertDictEqual(f.query, {})
#         f[1].value = ('x', 456.7)
#         self.assertDictEqual(f.query, {'$set': [123, ['x', 456.7]]})
#         f.pop(0)
#         self.assertDictEqual(f.query, {'$set': [['x', 456.7]]})
#         f.dirty_clear()
#         del f[0]
#         self.assertDictEqual(f.query, {'$set': []})
#
#     def test_tuple_query_1(self):
#         f = make_field((int, None))
#         self.assertDictEqual(f.query, {})
#         f[0].value = 123
#         self.assertDictEqual(f.query, {'$set': [123, None]})
#         f[1].value = 'foo'
#         self.assertDictEqual(f.query, {'$set': [123, 'foo']})
#         f.dirty_clear()
#         self.assertDictEqual(f.query, {})
#         del f[0]
#         self.assertDictEqual(f.query, {'$set': [None, 'foo']})
#
#     def test_list_tuple_query_1(self):
#         f = make_field((str, [None]))
#         self.assertDictEqual(f.query, {})
#         f.set_value(['foo', [1]])
#         self.assertDictEqual(f.query, {'$set': ['foo', [1]]})
#         f['1.2'].value = 'bar'
#         self.assertDictEqual(f.query, {'$set': ['foo', [1, None, 'bar']]})
#
#     def test_list_query_2(self):
#         f = make_field([{'a': int, 'b': None}])
#         self.assertDictEqual(f.query, {})
#         f.append_value({'a': 1})
#         self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}]})
#         f.append_value({'b': 'foo'})
#         self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}, {'a': None, 'b': 'foo'}]})
#         f[0].value = {'a': 123, 'b': 456.7}
#         self.assertDictEqual(f.query, {'$set': [{'a': 123, 'b': 456.7}, {'a': None, 'b': 'foo'}]})



#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2016, Asmodius
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of California, Berkeley nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


try:
    from setuptools import setup, find_packages
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup, find_packages

from mokito import version


setup(
    name='mokito',
    version=version,
    description="Mokito is an asynchronous ORM for working with MongoDB inside a Tornado app",
    long_description=open('README.md').read(),
    keywords='mongo mongodb tornado mokito',
    author='Asmodius',
    author_email='asmodius.a@gmail.com',
    license='New BSD License',
    url='https://github.com/Asmodius/mokito',
    download_url='https://github.com/Asmodius/mokito/archive/master.zip',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    zip_safe=False,
    packages=find_packages(),
    setup_requires=open("requirements.txt").read().split("\n"),
    install_requires=open("requirements.txt").read().split("\n"),
)



# import unittest
#
# from mokito.fields import IntField, FloatField, StringField
# from mokito.tools import make_field
# from mokito.collections import ListField, TupleField, DictField
#
#
# class TestCollectionFieldsDefault(unittest.TestCase):
#     def test_list_field_1(self):
#         f = make_field([123])
#         self.assertIsInstance(f, ListField)
#         self.assertIsInstance(f._rules, IntField)
#         self.assertListEqual(f.value, [123])
#         self.assertListEqual(f.self_value, [None])
#         self.assertEqual(f[0].value, 123)
#         self.assertListEqual(f.value, [123])
#         self.assertListEqual(f.self_value, [None])
#         f.append_value('111')
#         self.assertListEqual(f.value, [123, 111])
#         self.assertListEqual(f.self_value, [None, 111])
#         f.value = None
#         self.assertListEqual(f.value, [])
#         self.assertListEqual(f.self_value, [])
#
#     def test_tuple_field_1(self):
#         f = make_field((123, 'foo'))
#         self.assertIsInstance(f, TupleField)
#         self.assertIsInstance(f._val['0'], IntField)
#         self.assertTrue(f == f._val['0']._parent)
#         self.assertIsInstance(f._val['1'], StringField)
#         self.assertTrue(f == f._val['1']._parent)
#
#         self.assertListEqual(f.value, [123, 'foo'])
#         self.assertListEqual(f.self_value, [None, None])
#
#         f[0].value = 456
#         self.assertEqual(f.value, [456, 'foo'])
#         self.assertEqual(f.self_value, [456, None])
#         self.assertTrue(f == f[0]._parent)
#
#         f[1].value = 456
#         self.assertEqual(f.value, [456, '456'])
#         self.assertEqual(f.self_value, [456, '456'])
#         self.assertTrue(f == f[1]._parent)
#
#         f[0].value = None
#         self.assertEqual(f.value, [123, '456'])
#         self.assertEqual(f.self_value, [None, '456'])
#         self.assertTrue(f == f[0]._parent)
#
#         del f[1]
#         self.assertEqual(f.value, [123, 'foo'])
#         self.assertEqual(f.self_value, [None, None])
#         self.assertTrue(f == f[1]._parent)
#
#         f.clear()
#         self.assertListEqual(f.value, [123, 'foo'])
#         self.assertListEqual(f.self_value, [None, None])
#
#     def test_dict_field_1(self):
#         data = {
#             'f0': 'foo',
#             'f1': 123,
#             'f2': 456.7,
#         }
#         f = make_field(data)
#         self.assertIsInstance(f, DictField)
#         self.assertIsInstance(f._val['f0'], StringField)
#         self.assertTrue(f == f._val['f0']._parent)
#         self.assertIsInstance(f._val['f1'], IntField)
#         self.assertTrue(f == f._val['f1']._parent)
#         self.assertIsInstance(f._val['f2'], FloatField)
#         self.assertTrue(f == f._val['f2']._parent)
#
#         self.assertDictEqual(f.value, {'f0': 'foo', 'f1': 123, 'f2': 456.7})
#         f.value = {'f0': 'bar', 'f1': 456}
#         self.assertDictEqual(f.value, {'f0': 'bar', 'f1': 456, 'f2': 456.7})
#         f['f2'].value = 9
#         self.assertDictEqual(f.value, {'f0': 'bar', 'f1': 456, 'f2': 9.0})
#
#     def test_list_query_1(self):
#         f = make_field(['foo'])
#         self.assertDictEqual(f.query, {})
#         f.append_value('bar')
#         self.assertDictEqual(f.query, {'$set': ['foo', 'bar']})
#         del f[0]
#         self.assertDictEqual(f.query, {'$set': ['bar']})
#         f.dirty_clear()
#         self.assertDictEqual(f.query, {})
#         f[1].value = 'x'
#         self.assertDictEqual(f.query, {'$set': ['bar', 'x']})
#
#     def test_tuple_query_1(self):
#         f = make_field((123, str))
#         self.assertDictEqual(f.query, {})
#         f[1].value = 'foo'
#         self.assertDictEqual(f.query, {'$set': [123, 'foo']})
#         f.dirty_clear()
#         self.assertDictEqual(f.query, {})
#         del f[0]
#         self.assertDictEqual(f.query, {'$set': [123, 'foo']})
#         f.dirty_clear()
#         self.assertDictEqual(f.query, {})



class MokitoError(Exception):
    """Base class for all exceptions. """


# class MokitoDriverError(MokitoError):
#     pass
#
#
# class MokitoEmptyBulkError(MokitoDriverError):
#     message = "cannot do an empty bulk insert"
#
#
# class MokitoMasterChangedError(MokitoDriverError):
#     message = "master has changed"
#
#
# class MokitoInvalidURIError(MokitoDriverError):
#     message = "Invalid URI"
#
#
# class MokitoInvalidCursorError(MokitoDriverError):
#     message = "Cursor not valid at server"
#
#
# class MokitoParamError(MokitoDriverError):
#     message = "Invalid parameters"
#
#
# class MokitoConnectionError(MokitoDriverError):
#     pass
#
#
# class MokitoResponseError(MokitoDriverError):
#
#     def __init__(self, msg, code=None):
#         self.code = code
#         self.msg = msg
#
#     def __unicode__(self):
#         return u'Response error: %s code:%s' % (self.msg, self.code or '')
#
#     def __str__(self):
#         return str(self.__unicode__())


class MokitoORMError(MokitoError):
    pass


class MokitoDBREFError(MokitoORMError):
    pass


class MokitoChoiceError(MokitoORMError):
    pass



import copy
import inspect
import datetime

from bson import ObjectId
# from pymongo import ASCENDING, DESCENDING


SEPARATOR = '.'


def make_field(rules, _parent=None):
    from .models import Model
    # from .documents import Document
    from .fields import Field, AnyField, IntField, FloatField, StringField, BooleanField, ObjectIdField, DateTimeField
    from .collections import ListField, TupleField, DictField
    from .geo_fileds import GEOField, _Point1, _Point2

    if rules is None:
        return AnyField(_parent=_parent)

    elif isinstance(rules, Field):
        x = copy.deepcopy(rules)
        x._parent = _parent
        return x

    else:
        if inspect.isclass(rules):
            # print('l0-1')
            _type = rules
            _val = None
        else:
            # print('l0-2')
            _type = type(rules)
            _val = rules

        if _type is int:
            return IntField(_val, _parent=_parent)
        if _type is float:
            return FloatField(_val, _parent=_parent)
        if _type is str:
            return StringField(_val, _parent=_parent)
        if _type is bool:
            return BooleanField(_val, _parent=_parent)
        if _type is ObjectId:
            return ObjectIdField(_val, _parent=_parent)
        if _type is datetime.datetime:
            return DateTimeField(_val, _parent=_parent)

        if _type is list:
            # print('L1', rules, _val) # rules == _val - может rules поменять на _val ?
            return ListField(rules, _parent=_parent)
        if _type is tuple:
            return TupleField(rules, _parent=_parent)
        if _type is dict:
            return DictField(rules, _parent=_parent)

        # if issubclass(_type, Document):
        #     print('SC', _type)
        #     return _type(_parent=_parent)

        if issubclass(_type, _Point1):
            return _type(_parent=_parent)

        if issubclass(_type, _Point2):
            return _type(_parent=_parent)

        if issubclass(_type, GEOField):
            return _type(_parent=_parent)

        if issubclass(_type, Model):
            return _type(_parent=_parent)

        if issubclass(_type, Field):
            return _type(_parent=_parent)
            # x = _type(rules)
            # x._parent = _parent
            # return x
        raise TypeError()


# def fields2sort(sort):
#     if isinstance(sort, basestring):
#         sort = [sort]
#     return [(i[1:], DESCENDING,) if i[0] == '-' else
#             (i, ASCENDING,) if isinstance(i, basestring) else
#             i for i in sort]


# def norm_spec(spec):
#     if spec is None:
#         return {}
#
#     if isinstance(spec, basestring):
#         return {"_id": ObjectId(spec)}
#
#     if isinstance(spec, ObjectId):
#         return {"_id": spec}
#
#     return spec
